package com.navmii.sdksampleapp;

import android.Manifest;
import android.app.Activity;
import android.content.Intent;
import android.content.pm.PackageManager;
import android.os.Bundle;
import android.view.LayoutInflater;
import android.view.View;
import android.view.ViewGroup;
import android.widget.TextView;

import androidx.annotation.NonNull;
import androidx.annotation.StringRes;
import androidx.appcompat.app.AppCompatActivity;
import androidx.core.app.ActivityCompat;
import androidx.core.content.ContextCompat;
import androidx.recyclerview.widget.DividerItemDecoration;
import androidx.recyclerview.widget.LinearLayoutManager;
import androidx.recyclerview.widget.RecyclerView;

import com.navmii.sdksampleapp.geocodingsearch.GeocodingSearchActivity;
import com.navmii.sdksampleapp.geofence.GeofenceActivity;
import com.navmii.sdksampleapp.geoobjects.GeoObjectsActivity;
import com.navmii.sdksampleapp.map.MapActivity;
import com.navmii.sdksampleapp.navigation.NavigationActivity;
import com.navmii.sdksampleapp.places.PlacesActivity;
import com.navmii.sdksampleapp.poisearch.PoiSearchActivity;
import com.navmii.sdksampleapp.reversegeocoding.ReverseGeocodingActivity;

import java.util.Arrays;
import java.util.List;

public class MainActivity extends AppCompatActivity {

    private static final int REQUEST_CODE_PERMISSIONS = 1;

    private static final String[] REQUIRED_PERMISSIONS = {
            Manifest.permission.ACCESS_FINE_LOCATION
    };

    private static final List<ExampleItem> EXAMPLE_ITEMS = Arrays.asList(
            new ExampleItem(R.string.title_map, MapActivity.class),
            new ExampleItem(R.string.title_geocoding, GeocodingSearchActivity.class),
            new ExampleItem(R.string.title_reverseGeocoding, ReverseGeocodingActivity.class),
            new ExampleItem(R.string.title_poiSearch, PoiSearchActivity.class),
            new ExampleItem(R.string.title_navigation, NavigationActivity.class),
            new ExampleItem(R.string.title_places, PlacesActivity.class),
            new ExampleItem(R.string.title_geoObjects, GeoObjectsActivity.class),
            new ExampleItem(R.string.title_geofences, GeofenceActivity.class)
    );

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.activity_main);

        final RecyclerView recyclerView = findViewById(R.id.recycler_view);

        final ExampleItemsAdapter adapter =
                new ExampleItemsAdapter(EXAMPLE_ITEMS, new ExampleItemSelectedListener());

        final LinearLayoutManager layoutManager = new LinearLayoutManager(this);

        final RecyclerView.ItemDecoration itemDecoration =
                new DividerItemDecoration(this, layoutManager.getOrientation());

        recyclerView.setLayoutManager(layoutManager);
        recyclerView.addItemDecoration(itemDecoration);
        recyclerView.setAdapter(adapter);
    }

    @Override
    protected void onResume() {
        super.onResume();
        checkPermissions();
    }

    private void checkPermissions() {
        if (!hasPermissions()) {
            ActivityCompat.requestPermissions(
                    this,
                    REQUIRED_PERMISSIONS,
                    REQUEST_CODE_PERMISSIONS
            );
        }
    }

    private boolean hasPermissions() {
        for (String permission : REQUIRED_PERMISSIONS) {
            final int permissionCheck = ContextCompat.checkSelfPermission(this, permission);
            if (permissionCheck != PackageManager.PERMISSION_GRANTED) {
                return false;
            }
        }

        return true;
    }

    private class ExampleItemSelectedListener
            implements ExampleItemsAdapter.OnItemSelectedListener {

        @Override
        public void onItemSelected(ExampleItem selectedItem) {
            final Intent startIntent = new Intent(MainActivity.this, selectedItem.activityClass);
            startActivity(startIntent);
        }
    }

    // Item model
    private static final class ExampleItem {

        @StringRes
        private final int titleRes;
        private final Class<? extends Activity> activityClass;

        private ExampleItem(@StringRes int titleRes,
                            Class<? extends Activity> activityClass) {
            this.titleRes = titleRes;
            this.activityClass = activityClass;
        }
    }

    private static class ExampleItemViewHolder extends RecyclerView.ViewHolder
            implements View.OnClickListener {

        static ExampleItemViewHolder create(ViewGroup parent,
                                            OnItemSelectedListener onItemSelectedListener) {
            final View itemView = LayoutInflater.from(parent.getContext())
                    .inflate(R.layout.example_list_item, parent, false);

            return new ExampleItemViewHolder(itemView, onItemSelectedListener);
        }

        private final OnItemSelectedListener onItemSelectedListener;
        private final TextView titleTextView;

        private ExampleItemViewHolder(View itemView, OnItemSelectedListener onItemSelectedListener) {
            super(itemView);
            this.onItemSelectedListener = onItemSelectedListener;
            titleTextView = itemView.findViewById(R.id.text_title);
            itemView.setOnClickListener(this);
        }

        void onBind(ExampleItem item) {
            titleTextView.setText(item.titleRes);
        }

        @Override
        public void onClick(View v) {
            onItemSelectedListener.onItemSelected(getAdapterPosition());
        }

        interface OnItemSelectedListener {
            void onItemSelected(int adapterPosition);
        }
    }

    private static class ExampleItemsAdapter extends RecyclerView.Adapter<ExampleItemViewHolder>
            implements ExampleItemViewHolder.OnItemSelectedListener {

        private final List<ExampleItem> items;
        private final OnItemSelectedListener onItemSelectedListener;

        ExampleItemsAdapter(@NonNull List<ExampleItem> items,
                            @NonNull OnItemSelectedListener onItemSelectedListener) {
            this.items = items;
            this.onItemSelectedListener = onItemSelectedListener;
        }

        @Override
        public ExampleItemViewHolder onCreateViewHolder(ViewGroup parent, int viewType) {
            return ExampleItemViewHolder.create(parent, this);
        }

        @Override
        public void onBindViewHolder(ExampleItemViewHolder holder, int position) {
            holder.onBind(items.get(position));
        }

        @Override
        public int getItemCount() {
            return items.size();
        }

        @Override
        public void onItemSelected(int itemIndex) {
            onItemSelectedListener.onItemSelected(items.get(itemIndex));
        }

        interface OnItemSelectedListener {
            void onItemSelected(ExampleItem selectedItem);
        }
    }
}
