package com.navmii.sdksampleapp.external;

import android.util.JsonReader;

import androidx.annotation.Nullable;

import com.navmii.sdk.common.MapCoordinates;
import com.navmii.sdksampleapp.ExternalRouteParser;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.List;

public class SelectedInterventionsRouteParser implements ExternalRouteParser {
    private static final class Point {
        enum Type {
            SNAPPED,
            INTERPOLATED,
            FILTERED,
            RAW
        }

        private final MapCoordinates coordinates;
        private final Type type;
        private final boolean isEssential;

        Point(MapCoordinates coordinates, Type type, boolean isEssential) {
            this.coordinates = coordinates;
            this.type = type;
            this.isEssential = isEssential;
        }
    }

    @Override
    public Result parse(InputStream inputStream) throws ParseException {
        List<Point> points = new ArrayList<>();

        JsonReader jsonReader =
                new JsonReader(new BufferedReader(new InputStreamReader(inputStream)));

        try {
            jsonReader.beginObject();

            boolean pointsFound = false;

            while (!pointsFound && jsonReader.hasNext()) {
                switch (jsonReader.nextName()) {
                    case "points":
                        pointsFound = true;
                        break;
                    default:
                        jsonReader.skipValue();
                        break;
                }
            }

            if (!pointsFound) {
                throw new ParseException("Object 'points' not found");
            }

            jsonReader.beginArray();
            while (jsonReader.hasNext()) {
                points.add(readPoint(jsonReader));
            }
            jsonReader.endArray();

            jsonReader.endObject();
        } catch (IOException e) {
            throw new ParseException(e.getMessage());
        }

        if (points.size() < 2) {
            throw new ParseException("Too few points: " + points.size());
        }

        List<MapCoordinates> meaningfulPoints = new ArrayList<>();
        List<Integer> waypointIndices = new ArrayList<>();

        addMeaningfulPoint(meaningfulPoints, points.get(0));

        for (int i = 1, size = points.size() - 1; i < size; ++i) {
            Point point = points.get(i);

            if (addMeaningfulPoint(meaningfulPoints, point)) {
                if (point.isEssential) {
                    waypointIndices.add(meaningfulPoints.size() - 1);
                }
            }
        }

        addMeaningfulPoint(meaningfulPoints, points.get(points.size() - 1));

        int[] waypointIndicesArray = new int[waypointIndices.size()];
        for (int i = 0, size = waypointIndices.size(); i < size; ++i) {
            waypointIndicesArray[i] = waypointIndices.get(i);
        }

        return new Result(
                meaningfulPoints.toArray(new MapCoordinates[meaningfulPoints.size()]),
                waypointIndicesArray);
    }

    private boolean addMeaningfulPoint(List<MapCoordinates> meaningfulPoints, Point point) {
        if (isMeaningfulPoint(point)) {
            meaningfulPoints.add(point.coordinates);
            return true;
        }

        return false;
    }

    private Point readPoint(JsonReader jsonReader) throws IOException {
        MapCoordinates coordinates = null;
        Point.Type type = null;
        boolean isEssential = false;

        jsonReader.beginObject();
        while (jsonReader.hasNext()) {
            switch (jsonReader.nextName()) {
                case "location":
                    jsonReader.beginArray();
                    double lat = jsonReader.nextDouble();
                    double lon = jsonReader.nextDouble();
                    coordinates = new MapCoordinates(lat, lon);
                    jsonReader.endArray();
                    break;
                case "type":
                    type = typeFromString(jsonReader.nextString());
                    break;
                case "isEssential":
                    isEssential = jsonReader.nextBoolean();
                    break;
                default:
                    jsonReader.skipValue();
                    break;
            }
        }
        jsonReader.endObject();

        if (!isValidCoordinates(coordinates)) {
            throw new ParseException("Invalid coordinates: " + coordinates);
        }

        if (type == null) {
            throw new ParseException("Invalid type: " + null);
        }

        return new Point(coordinates, type, isEssential);
    }

    private static Point.Type typeFromString(String typeString) {
        switch (typeString) {
            case "Snapped":
                return Point.Type.SNAPPED;
            case "Interpolated":
                return Point.Type.INTERPOLATED;
            case "Filtered":
                return Point.Type.FILTERED;
            case "Raw":
                return Point.Type.RAW;
        }

        return null;
    }

    private boolean isMeaningfulPoint(Point point) {
        return point.isEssential ||
                point.type == Point.Type.SNAPPED ||
                point.type == Point.Type.INTERPOLATED;
    }

    private boolean isValidCoordinates(@Nullable MapCoordinates coordinates) {
        return coordinates != null &&
                (coordinates.getLatitude() != 0 || coordinates.getLongitude() != 0);
    }
}
