package com.navmii.sdksampleapp.geocodingsearch;

import android.app.ProgressDialog;
import android.content.Intent;
import android.graphics.Color;
import android.graphics.PointF;
import android.os.Bundle;
import android.os.Handler;
import android.util.Log;
import android.view.MenuItem;
import android.view.View;
import android.view.WindowManager;
import android.widget.SearchView;
import android.widget.Toast;

import androidx.annotation.Nullable;
import androidx.appcompat.app.AppCompatActivity;
import androidx.appcompat.widget.Toolbar;
import androidx.core.content.res.ResourcesCompat;

import com.navmii.sdk.Sdk;
import com.navmii.sdk.common.MapCoordinates;
import com.navmii.sdk.common.MapRectangle;
import com.navmii.sdk.common.Place;
import com.navmii.sdk.map.CameraPosition;
import com.navmii.sdk.map.GeoMarker;
import com.navmii.sdk.map.GeoPolyline;
import com.navmii.sdk.map.MapView;
import com.navmii.sdk.positioning.Position;
import com.navmii.sdk.search.GeocodingRequest;
import com.navmii.sdk.search.Request;
import com.navmii.sdk.search.RequestError;
import com.navmii.sdksampleapp.CommonPreferenceActivity;
import com.navmii.sdksampleapp.PreferenceUtils;
import com.navmii.sdksampleapp.R;
import com.navmii.sdksampleapp.SearchResultsActivity;
import com.navmii.sdksampleapp.sdk.AddressInfo;
import com.navmii.sdksampleapp.sdk.PlaceInfo;
import com.navmii.sdksampleapp.sdk.SdkFragment;

import java.io.File;
import java.util.ArrayList;
import java.util.List;

public class GeocodingSearchActivity extends AppCompatActivity
        implements View.OnClickListener, SearchView.OnQueryTextListener {

    private static final int REQUEST_CODE_SEARCH_RESULT = 1;

    private final Sdk sdk = Sdk.getInstance();
    private final Sdk.StateChangeListener stateChangeListener = new SdkStateListener();

    private MapView mapView;
    private View initializationProgressBar;
    private File flag;
    private SearchView searchView;

    private GeoMarker searchRectangleTopLeftMarker;
    private GeoMarker searchRectangleBottomRightMarker;
    private GeoPolyline searchRectanglePolyline;

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.activity_geocoding_search);
        getWindow().addFlags(WindowManager.LayoutParams.FLAG_KEEP_SCREEN_ON);

        final Toolbar toolbar = findViewById(R.id.toolbar);
        searchView = toolbar.findViewById(R.id.search_bar);
        setSupportActionBar(toolbar);

        if (getSupportActionBar() != null) {
            getSupportActionBar().setDisplayHomeAsUpEnabled(true);
            getSupportActionBar().setDisplayShowHomeEnabled(true);
        }

        if (getFragmentManager().findFragmentByTag(SdkFragment.SDK_FRAGMENT_TAG) == null) {
            getFragmentManager()
                    .beginTransaction()
                    .add(new SdkFragment(), SdkFragment.SDK_FRAGMENT_TAG)
                    .commit();
        }

        mapView = findViewById(R.id.map_view);
        mapView.addMapTapListener(new MapTapListener());
        searchView.setOnQueryTextListener(this);
        initializationProgressBar = findViewById(R.id.initialization_progressbar);
        findViewById(R.id.settings_button).setOnClickListener(this);
    }

    @Override
    protected void onStart() {
        super.onStart();

        if (sdk.getState() == Sdk.State.INITIALIZED) {
            sdk.setActiveMapView(mapView);
        }

        sdk.addStateChangeListener(stateChangeListener);

        updateControlsVisibility();
    }

    @Override
    protected void onStop() {
        super.onStop();

        if (sdk.getState() == Sdk.State.INITIALIZED) {
            sdk.setActiveMapView(null);
        }

        sdk.removeStateChangeListener(stateChangeListener);
    }

    @Override
    public boolean onOptionsItemSelected(MenuItem item) {
        switch (item.getItemId()) {
            case android.R.id.home:
                finish();
                return true;
        }

        return super.onOptionsItemSelected(item);
    }

    private class SdkStateListener implements Sdk.StateChangeListener {
        @Override
        public void onStateChanged(Sdk.State state) {
            if (state == Sdk.State.INITIALIZED) {
                sdk.setActiveMapView(mapView);
            }

            updateControlsVisibility();
        }
    }

    @Override
    public void onClick(View view) {
        switch (view.getId()) {
            case R.id.settings_button:
                startGeocodingPreferenceActivity();
                break;
        }
    }

    private void updateControlsVisibility() {
        initializationProgressBar.setVisibility(
                sdk.getState() == Sdk.State.INITIALIZING ? View.VISIBLE : View.GONE);

        searchView.setVisibility(
                sdk.getState() == Sdk.State.INITIALIZED ? View.VISIBLE : View.GONE);

        findViewById(R.id.settings_button).setVisibility(
                sdk.getState() == Sdk.State.INITIALIZED ? View.VISIBLE : View.GONE);
    }

    private void startGeoCodingSearch(String query) {
        if (sdk.getState() != Sdk.State.INITIALIZED) {
            return;
        }

        final Position currentPosition = sdk.getPositionManager().getCurrentPosition();
        if (currentPosition == null) {
            Toast.makeText(this, R.string.no_current_position, Toast.LENGTH_SHORT).show();
            return;
        }

        final String maxResultCountString = PreferenceUtils.getString(
                this,
                R.string.key_geocodingSearch_maxResultCount,
                R.string.defaultValue_geocodingSearch_maxResultCount
        );

        final int maxResultCount = Integer.parseInt(maxResultCountString);

        final GeocodingRequest.Builder builder = sdk.getSearchManager()
                .createGeocodingRequest(query, currentPosition.coordinates())
                .setMaxResultCount(maxResultCount);

        final boolean searchWithinVisibleArea = PreferenceUtils.getBoolean(
                this,
                R.string.key_geocodingSearch_searchWithinVisibleArea,
                R.bool.defaultValue_geocodingSearch_searchWithinVisibleArea);

        if (searchWithinVisibleArea) {
            MapView.CameraController cameraController = mapView.getCameraController();
            builder.setViewBox(cameraController.getMapRectangleForCameraPosition(
                    cameraController.getCameraPosition()));
        } else {
            final int searchRadius = PreferenceUtils.getInteger(
                    this,
                    R.string.key_geocodingSearch_searchRadius,
                    R.integer.defaultValue_geocodingSearch_searchRadius);
            builder.setSearchRadius(searchRadius);
        }

        final Request geocodingRequest = builder.build();

        final ProgressDialog progressDialog = ProgressDialog.show(
                this,
                getString(R.string.dialog_loading),
                getString(R.string.dialog_wait),
                true
        );

        geocodingRequest.start(new Request.CompletionListener() {
            @Override
            public void onRequestCompleted(Request request, List<Place> results) {
                progressDialog.dismiss();
                startSearchResultsActivity(currentPosition.coordinates(), results);
            }

            @Override
            public void onRequestFailed(Request request, RequestError error) {
                Log.i("GeocodingRequest", "Request failed: " + error);
            }
        });
    }

    private void startSearchResultsActivity(MapCoordinates searchCoordinates, List<Place> places) {
        final ArrayList<PlaceInfo> placeInfos = new ArrayList<>();
        for (Place place : places) {
            placeInfos.add(
                    new PlaceInfo(
                            place.getName(),
                            "",
                            new AddressInfo(place.getAddress(), place.getCoordinates())));
        }

        final Intent startIntent = SearchResultsActivity.createStartIntent(
                this, searchCoordinates, placeInfos);
        startActivityForResult(startIntent, REQUEST_CODE_SEARCH_RESULT);
    }

    @Override
    public boolean onQueryTextSubmit(String query) {
        startGeoCodingSearch(query);
        return false;
    }

    @Override
    public boolean onQueryTextChange(String newText) {
        return false;
    }

    @Override
    protected void onActivityResult(int requestCode, int resultCode, final Intent data) {
        super.onActivityResult(requestCode, resultCode, data);

        if (requestCode == REQUEST_CODE_SEARCH_RESULT && resultCode == RESULT_OK) {

            // onActivityResult may be called before onStart,
            // when it's prohibited to used some of MapView functionality.
            new Handler().post(new Runnable() {
                @Override
                public void run() {
                    MapCoordinates mapCoordinates = data.getParcelableExtra(
                            SearchResultsActivity.EXTRA_RESULT_MAP_COORDINATES);

                    moveCamera(mapCoordinates);
                    mapView.removeAllGeoObjects();
                    mapView.setFollowGpsPosition(false);
                    addMarker(mapCoordinates);
                }
            });
        }
    }

    private void moveCamera(MapCoordinates mapCoordinates) {
        CameraPosition cameraPosition =
                new CameraPosition.Builder(mapView.getCameraController().getCameraPosition())
                        .setTargetLocation(mapCoordinates)
                        .build();
        mapView.getCameraController().animateCamera(cameraPosition);
    }

    private void addMarker(MapCoordinates mapCoordinates) {
        GeoMarker marker = new GeoMarker(mapCoordinates, getFlagFile(), getFlagAnchorPoint());
        marker.setClickable(false);
        mapView.addGeoObject(marker);
    }

    @Nullable
    private File getFlagFile() {
        if (flag == null) {
            flag = new File(generateFlagPath());
        }

        return flag;
    }

    private String generateFlagPath() {
        String fileDirs = this.getFilesDir().getAbsolutePath();
        return fileDirs + File.separator + "assets" + File.separator + "flag_green.png";
    }

    private PointF getFlagAnchorPoint() {
        return new PointF(0.3f, 0.9f);
    }

    private void startGeocodingPreferenceActivity() {
        final Intent startIntent = CommonPreferenceActivity.createStartIntent(
                this,
                R.xml.preferences_geocoding_search,
                R.string.title_geocodingPreferences
        );

        startActivity(startIntent);
    }

    private GeoMarker createSearchRectangleMarker(MapCoordinates coordinates) {
        return new GeoMarker(coordinates,
                ResourcesCompat.getDrawable(getResources(), R.drawable.map_marker_1, getTheme()),
                new PointF(0.5f, 0.5f));
    }

    private class MapTapListener implements MapView.MapTapListener {
        @Override
        public void onSingleTap(MapCoordinates coordinates) {
            if (searchRectanglePolyline != null) {
                mapView.removeGeoObject(searchRectanglePolyline);
                mapView.removeGeoObject(searchRectangleTopLeftMarker);
                mapView.removeGeoObject(searchRectangleBottomRightMarker);
                searchRectanglePolyline = null;
                searchRectangleTopLeftMarker = null;
                searchRectangleBottomRightMarker = null;
                return;
            }

            if (searchRectangleTopLeftMarker == null) {
                searchRectangleTopLeftMarker = createSearchRectangleMarker(coordinates);
                mapView.addGeoObject(searchRectangleTopLeftMarker);
            } else {
                searchRectangleBottomRightMarker = createSearchRectangleMarker(coordinates);
                mapView.addGeoObject(searchRectangleBottomRightMarker);

                searchRectanglePolyline = new GeoPolyline(new MapRectangle(
                        searchRectangleTopLeftMarker.getCoordinates(),
                        searchRectangleBottomRightMarker.getCoordinates()));
                searchRectanglePolyline.setWidth(4);
                searchRectanglePolyline.setColor(Color.BLUE);
                mapView.addGeoObject(searchRectanglePolyline);
            }
        }

        @Override
        public void onDoubleTap(MapCoordinates coordinates) {
        }

        @Override
        public void onLongTap(MapCoordinates coordinates) {
        }
    }
}
