package com.navmii.sdksampleapp.geofence;

import android.annotation.SuppressLint;
import android.content.Intent;
import android.content.SharedPreferences;
import android.graphics.Color;
import android.os.Bundle;
import android.preference.PreferenceManager;
import android.view.Menu;
import android.view.MenuInflater;
import android.view.MenuItem;
import android.view.View;
import android.view.WindowManager;

import androidx.annotation.ColorInt;
import androidx.annotation.StringRes;
import androidx.appcompat.app.AppCompatActivity;

import com.google.android.material.snackbar.Snackbar;
import com.navmii.sdk.Sdk;
import com.navmii.sdk.common.MapCoordinates;
import com.navmii.sdk.geofence.Geofence;
import com.navmii.sdk.geofence.GeofenceCircular;
import com.navmii.sdk.geofence.GeofenceManager;
import com.navmii.sdk.geofence.GeofencePolygonal;
import com.navmii.sdk.map.GeoCircle;
import com.navmii.sdk.map.GeoObject;
import com.navmii.sdk.map.GeoPolygon;
import com.navmii.sdk.map.MapView;
import com.navmii.sdksampleapp.CommonPreferenceActivity;
import com.navmii.sdksampleapp.PreferenceChangeListener;
import com.navmii.sdksampleapp.PreferenceUtils;
import com.navmii.sdksampleapp.R;
import com.navmii.sdksampleapp.sdk.SdkFragment;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

public class GeofenceActivity extends AppCompatActivity implements View.OnClickListener {

    private final Sdk sdk = Sdk.getInstance();
    private final Sdk.StateChangeListener stateChangeListener = new SdkStateListener();

    private final MapView.MapTapListener mapTapListener = new MapTapListener();

    private SharedPreferences.OnSharedPreferenceChangeListener sharedPreferenceChangeListener;

    private final List<GeoObject> geoObjects = new ArrayList<GeoObject>();
    private GeoObject currentGeoObject = null;
    private GeoType geoType = GeoType.None;
    private long radius = 120;

    private MapView mapView;
    private View initializationProgressBar;
    private View settings;
    private MenuItem addGeofenceItem;
    private MenuItem doneGeofenceItem;
    private MenuItem cancelGeofenceItem;
    private MenuItem clearAllItem;
    private boolean isGeofenceAdding = false;

    private final GeofenceListener geofenceListener = new GeofenceListener();

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.activity_geofences);
        getWindow().addFlags(WindowManager.LayoutParams.FLAG_KEEP_SCREEN_ON);

        if (getSupportActionBar() != null) {
            getSupportActionBar().setDisplayHomeAsUpEnabled(true);
            getSupportActionBar().setDisplayShowHomeEnabled(true);
        }

        if (getFragmentManager().findFragmentByTag(SdkFragment.SDK_FRAGMENT_TAG) == null) {
            getFragmentManager()
                    .beginTransaction()
                    .add(new SdkFragment(), SdkFragment.SDK_FRAGMENT_TAG)
                    .commit();
        }

        sharedPreferenceChangeListener = PreferenceChangeListener.create(
                this,
                PREFERENCE_KEY_IDS_TO_LISTEN,
                new PreferenceChangeCallback()
        );

        mapView = findViewById(R.id.map_view);
        initializationProgressBar = findViewById(R.id.initialization_progressbar);
        settings = findViewById(R.id.settings_button);
        settings.setOnClickListener(this);
    }

    @Override
    protected void onStart() {
        super.onStart();

        if (sdk.getState() == Sdk.State.INITIALIZED) {
            sdk.setActiveMapView(mapView);
            sdk.getGeofenceManager().addGeofenceListener(geofenceListener);

            applyAllGeoObjectPreferences();
        }

        sdk.addStateChangeListener(stateChangeListener);

        updateControlsVisibility();

        mapView.addMapTapListener(mapTapListener);

        PreferenceManager.getDefaultSharedPreferences(this)
                .registerOnSharedPreferenceChangeListener(sharedPreferenceChangeListener);
    }

    @Override
    protected void onStop() {
        super.onStop();

        if (sdk.getState() == Sdk.State.INITIALIZED) {
            sdk.setActiveMapView(null);
            sdk.getGeofenceManager().removeGeofenceListener(geofenceListener);
        }

        sdk.removeStateChangeListener(stateChangeListener);

        mapView.removeMapTapListener(mapTapListener);

        PreferenceManager.getDefaultSharedPreferences(this)
                .unregisterOnSharedPreferenceChangeListener(sharedPreferenceChangeListener);
    }

    @Override
    public boolean onCreateOptionsMenu(Menu menu) {
        MenuInflater inflater = getMenuInflater();
        inflater.inflate(R.menu.menu_geofence, menu);
        addGeofenceItem = menu.findItem(R.id.add_geofence);
        doneGeofenceItem = menu.findItem(R.id.done_geofence);
        cancelGeofenceItem = menu.findItem(R.id.cancel_geofence);
        clearAllItem = menu.findItem(R.id.clear_all_geofences);
        return true;
    }

    @SuppressLint("NonConstantResourceId")
    @Override
    public boolean onOptionsItemSelected(MenuItem item) {
        switch (item.getItemId()) {
            case android.R.id.home:
                finish();
                return true;
            case R.id.add_geofence: {
                isGeofenceAdding = true;
                applyAllGeoObjectPreferences();
                updateMenuItems();
                return true;
            }
            case R.id.done_geofence: {
                isGeofenceAdding = false;
                if(geoType == GeoType.Polygonal) addGeofence();
                updateMenuItems();
                currentGeoObject = null;
                return true;
            }
            case R.id.cancel_geofence: {
                isGeofenceAdding = false;
                if(currentGeoObject != null) {
                    mapView.removeGeoObject(currentGeoObject);
                    currentGeoObject = null;
                }
                updateMenuItems();
                return true;
            }
            case R.id.clear_all_geofences: {
                mapView.removeAllGeoObjects();
                geoObjects.clear();
                GeofenceManager getGeofenceManager = sdk.getGeofenceManager();
                if (getGeofenceManager != null) {
                    getGeofenceManager.clearAllGeofence();
                }
                updateMenuItems();
                return true;
            }
            default:
                return false;
        }
    }

    private void updateMenuItems() {
        addGeofenceItem.setVisible(!isGeofenceAdding);
        doneGeofenceItem.setVisible(isGeofenceAdding);
        cancelGeofenceItem.setVisible(isGeofenceAdding);
        clearAllItem.setVisible(!isGeofenceAdding && !geoObjects.isEmpty());
    }

    private void addGeofence() {
        GeofenceManager getGeofenceManager = sdk.getGeofenceManager();
        if (getGeofenceManager == null) {
            return;
        }
        if(currentGeoObject instanceof GeoPolygon) {
            Geofence geofence = new GeofencePolygonal(((GeoPolygon)currentGeoObject).getVertices());
            geofence.setTag("tag_polygon_" + geoObjects.size());
            getGeofenceManager.addGeofence(geofence);
        } else if (currentGeoObject instanceof GeoCircle) {
            Geofence geofence = new GeofenceCircular(((GeoCircle)currentGeoObject).getCenter(), radius);
            geofence.setTag("tag_circle_" + geoObjects.size());
            getGeofenceManager.addGeofence(geofence);
        }
        geoObjects.add(currentGeoObject);
    }

    private class SdkStateListener implements Sdk.StateChangeListener {
        @Override
        public void onStateChanged(Sdk.State state) {
            if (state == Sdk.State.INITIALIZED) {
                sdk.setActiveMapView(mapView);
                applyAllGeoObjectPreferences();
                mapView.setMapZoomOnDoubleTapEnabled(false);
            }
            updateControlsVisibility();
        }
    }

    private class GeofenceListener implements GeofenceManager.GeofenceListener {

        @Override
        public void onGeofenceEnter(String id, String tag) {
            Snackbar.make(mapView,
                    "Enter to Geofence. tag: " + tag,
                    3000
            ).show();
        }

        @Override
        public void onGeofenceExit(String id, String tag) {
            Snackbar.make(mapView,
                    "Exit Geofence. tag: " + tag,
                    3000
            ).show();
        }
    }

    @Override
    public void onClick(View view) {
        switch (view.getId()) {
            case R.id.settings_button:
                startGeofencePreferenceActivity();
                break;
        }
    }

    private void updateControlsVisibility() {
        settings.setVisibility(
                sdk.getState() == Sdk.State.INITIALIZED ? View.VISIBLE : View.GONE);

        initializationProgressBar.setVisibility(
                sdk.getState() == Sdk.State.INITIALIZING ? View.VISIBLE : View.GONE);
    }

    private void startGeofencePreferenceActivity() {
        final Intent startIntent = CommonPreferenceActivity.createStartIntent(
                this, R.xml.preferences_geofence, R.string.title_geofencePreferences
        );

        startActivity(startIntent);
    }

    private class MapTapListener implements MapView.MapTapListener {
        @Override
        public void onSingleTap(MapCoordinates mapCoordinates) {
            if (!isGeofenceAdding) return;
            switch (geoType) {
                case Polygonal: {
                    if(currentGeoObject == null) {
                        currentGeoObject = new GeoPolygon(new ArrayList<>());
                        applyAllPreferencesForCurrentGeoObject();
                    }
                    ((GeoPolygon)currentGeoObject).addVertex(mapCoordinates);
                }
                break;
                case Circular: {
                    currentGeoObject = new GeoCircle(
                            mapCoordinates, radius,
                            getFillColorPreference(), getStrokeColorPreference(),
                            getStrokeWidthPreference());
                    addGeofence();
                }
                break;
            }
            mapView.addGeoObject(currentGeoObject);
        }

        @Override
        public void onDoubleTap(MapCoordinates mapCoordinates) {
        }

        @Override
        public void onLongTap(MapCoordinates mapCoordinates) {
        }
    }

    // region Preferences
    private static final List<Integer> PREFERENCE_KEY_IDS_TO_LISTEN = Arrays.asList(
            R.string.key_geofence_fillColor,
            R.string.key_geofence_fillColorAlpha,
            R.string.key_geofence_strokeColor,
            R.string.key_geofence_strokeWidth,
            R.string.key_geofence_type
    );

    private class PreferenceChangeCallback implements PreferenceChangeListener.Callback {
        @SuppressLint("NonConstantResourceId")
        @Override
        public void onPreferenceChanged(@StringRes int keyId) {
            switch (keyId) {
                case R.string.key_geofence_fillColor:
                    applyFillColorPreference();
                    break;
                case R.string.key_geofence_fillColorAlpha:
                    applyFillColorPreference();
                    break;
                case R.string.key_geofence_strokeColor:
                    applyStrokeColorPreference();
                    break;
                case R.string.key_geofence_strokeWidth:
                    applyStrokeWidthPreference();
                    break;
                case R.string.key_geofence_type:
                    applyGeofenceTypePreference();
                    break;
            }
        }
    }

    private void applyAllGeoObjectPreferences() {
        applyFillColorPreference();
        applyStrokeColorPreference();
        applyStrokeWidthPreference();
        applyGeofenceTypePreference();
    }

    private void applyAllPreferencesForCurrentGeoObject() {
        applyFillColorForCurrentGeoObject();
        applyStrokeColorForCurrentGeoObject();
        applyStrokeWidthForCurrentGeoObject();
    }

    private GeoType getGeofenceTypePreference() {
        String value = PreferenceUtils.getString(
                this,
                R.string.key_geofence_type,
                R.string.defaultValue_entryValues_geofence_type
        );
        GeoType type = GeoType.None;
        switch (value) {
            case "polygonal" :
                type = GeoType.Polygonal;
            break;
            case "circular" :
                type = GeoType.Circular;
                break;
        }
        return type;
    }

    private @ColorInt int getFillColorPreference() {
        String alpha = PreferenceUtils.getString(
                this,
                R.string.key_geofence_fillColorAlpha,
                R.string.defaultValue_entryValues_geofence_fillColorAlpha
        );
        String value = PreferenceUtils.getString(
                this,
                R.string.key_geofence_fillColor,
                R.string.defaultValue_geoObjects_polygonFillColor
        ).replace("#", alpha);
        return Color.parseColor(value);
    }

    private @ColorInt int getStrokeColorPreference() {
        String value = PreferenceUtils.getString(
                this,
                R.string.key_geofence_strokeColor,
                R.string.defaultValue_geoObjects_polygonStrokeColor
        );
        return Color.parseColor(value);
    }

    private float getStrokeWidthPreference() {
        String value = PreferenceUtils.getString(
                this,
                R.string.key_geofence_strokeWidth,
                R.string.defaultValue_geoObjects_polygonStrokeWidth
        );
        return Float.parseFloat(value);
    }

    private void applyGeofenceTypePreference() {
        currentGeoObject = null;
        geoType = getGeofenceTypePreference();
    }

    private void applyFillColorPreference() {
        int color = getFillColorPreference();
        for (GeoObject geoObject : geoObjects) {
            if(geoObject instanceof GeoPolygon) {
                ((GeoPolygon) geoObject).setFillColor(color);
            }
        }
    }

    private void applyStrokeColorPreference() {
        int color = getStrokeColorPreference();
        for (GeoObject geoObject : geoObjects) {
            if(geoObject instanceof GeoPolygon) {
                ((GeoPolygon) geoObject).setStrokeColor(color);
            }
        }
    }

    private void applyStrokeWidthPreference() {
        float strokeWidth = getStrokeWidthPreference();
        for (GeoObject geoObject : geoObjects) {
            if(geoObject instanceof GeoPolygon) {
                ((GeoPolygon) geoObject).setStrokeWidth(strokeWidth);
            }
        }
    }

    private void applyFillColorForCurrentGeoObject() {
        int color = getFillColorPreference();
        if(currentGeoObject instanceof GeoPolygon) {
            ((GeoPolygon) currentGeoObject).setFillColor(color);
        }
    }

    private void applyStrokeColorForCurrentGeoObject() {
        int color = getStrokeColorPreference();
        if(currentGeoObject instanceof GeoPolygon) {
            ((GeoPolygon) currentGeoObject).setStrokeColor(color);
        }
    }

    private void applyStrokeWidthForCurrentGeoObject() {
        float strokeWidth = getStrokeWidthPreference();
        if(currentGeoObject instanceof GeoPolygon) {
            ((GeoPolygon) currentGeoObject).setStrokeWidth(strokeWidth);
        }
    }

    // endregion
}

enum GeoType {
    None, Polygonal, Circular
}