package com.navmii.sdksampleapp.geoobjects;

import android.content.Intent;
import android.content.SharedPreferences;
import android.graphics.Color;
import android.graphics.PointF;
import android.graphics.drawable.Drawable;
import android.os.Bundle;
import android.preference.PreferenceManager;

import androidx.annotation.StringRes;
import androidx.appcompat.app.AppCompatActivity;
import androidx.core.content.ContextCompat;
import android.view.Menu;
import android.view.MenuInflater;
import android.view.MenuItem;
import android.view.View;
import android.view.WindowManager;
import android.widget.Toast;

import com.navmii.sdk.Sdk;
import com.navmii.sdk.common.MapCoordinates;
import com.navmii.sdk.common.MapRectangle;
import com.navmii.sdk.map.CameraPosition;
import com.navmii.sdk.map.GeoMarker;
import com.navmii.sdk.map.GeoObject;
import com.navmii.sdk.map.GeoPolygon;
import com.navmii.sdk.map.GeoPolyline;
import com.navmii.sdk.map.MapView;
import com.navmii.sdksampleapp.CommonPreferenceActivity;
import com.navmii.sdksampleapp.PreferenceChangeListener;
import com.navmii.sdksampleapp.PreferenceUtils;
import com.navmii.sdksampleapp.R;
import com.navmii.sdksampleapp.sdk.SdkFragment;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

public class GeoObjectsActivity extends AppCompatActivity implements View.OnClickListener {

    private final Sdk sdk = Sdk.getInstance();
    private final Sdk.StateChangeListener stateChangeListener = new SdkStateListener();

    private final MapView.MapTapListener mapTapListener = new MapTapListener();
    private final MapView.GeoObjectListener geoObjectListener = new GeoObjectListener();

    private SharedPreferences.OnSharedPreferenceChangeListener sharedPreferenceChangeListener;

    private final GeoPolyline geoPolyline = new GeoPolyline(new ArrayList<MapCoordinates>());

    private final GeoPolygon geoPolygon = new GeoPolygon(new ArrayList<MapCoordinates>());

    private MapView mapView;
    private View initializationProgressBar;
    private View settings;

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.activity_geo_objects);
        getWindow().addFlags(WindowManager.LayoutParams.FLAG_KEEP_SCREEN_ON);

        if (getSupportActionBar() != null) {
            getSupportActionBar().setDisplayHomeAsUpEnabled(true);
            getSupportActionBar().setDisplayShowHomeEnabled(true);
        }

        if (getFragmentManager().findFragmentByTag(SdkFragment.SDK_FRAGMENT_TAG) == null) {
            getFragmentManager()
                    .beginTransaction()
                    .add(new SdkFragment(), SdkFragment.SDK_FRAGMENT_TAG)
                    .commit();
        }

        sharedPreferenceChangeListener = PreferenceChangeListener.create(
                this,
                PREFERENCE_KEY_IDS_TO_LISTEN,
                new PreferenceChangeCallback()
        );

        mapView = findViewById(R.id.map_view);
        initializationProgressBar = findViewById(R.id.initialization_progressbar);
        settings = findViewById(R.id.settings_button);
        settings.setOnClickListener(this);
    }

    @Override
    protected void onStart() {
        super.onStart();

        if (sdk.getState() == Sdk.State.INITIALIZED) {
            sdk.setActiveMapView(mapView);
            applyAllGeoObjectPreferences();
        }

        sdk.addStateChangeListener(stateChangeListener);

        updateControlsVisibility();

        mapView.addMapTapListener(mapTapListener);
        mapView.addGeoObjectListener(geoObjectListener);

        PreferenceManager.getDefaultSharedPreferences(this)
                .registerOnSharedPreferenceChangeListener(sharedPreferenceChangeListener);
    }

    @Override
    protected void onStop() {
        super.onStop();

        if (sdk.getState() == Sdk.State.INITIALIZED) {
            sdk.setActiveMapView(null);
        }

        sdk.removeStateChangeListener(stateChangeListener);

        mapView.removeMapTapListener(mapTapListener);
        mapView.removeGeoObjectListener(geoObjectListener);

        PreferenceManager.getDefaultSharedPreferences(this)
                .unregisterOnSharedPreferenceChangeListener(sharedPreferenceChangeListener);
    }

    @Override
    public boolean onCreateOptionsMenu(Menu menu) {
        MenuInflater inflater = getMenuInflater();
        inflater.inflate(R.menu.menu_geo_objects, menu);
        return true;
    }

    @Override
    public boolean onOptionsItemSelected(MenuItem item) {
        switch (item.getItemId()) {
            case android.R.id.home:
                finish();
                return true;
            case R.id.remove_all_geo_objects:
                mapView.removeAllGeoObjects();
                geoPolygon.removeAllVertices();
                geoPolyline.removeAllVertices();
                return true;
        }

        return super.onOptionsItemSelected(item);
    }

    private class SdkStateListener implements Sdk.StateChangeListener {
        @Override
        public void onStateChanged(Sdk.State state) {
            if (state == Sdk.State.INITIALIZED) {
                sdk.setActiveMapView(mapView);
                applyAllGeoObjectPreferences();
                mapView.setMapZoomOnDoubleTapEnabled(false);
            }
            updateControlsVisibility();
        }
    }

    @Override
    public void onClick(View view) {
        switch (view.getId()) {
            case R.id.settings_button:
                startGeoObjectsPreferenceActivity();
                break;
        }
    }

    private void updateControlsVisibility() {
        settings.setVisibility(
                sdk.getState() == Sdk.State.INITIALIZED ? View.VISIBLE : View.GONE);

        initializationProgressBar.setVisibility(
                sdk.getState() == Sdk.State.INITIALIZING ? View.VISIBLE : View.GONE);
    }

    private void startGeoObjectsPreferenceActivity() {
        final Intent startIntent = CommonPreferenceActivity.createStartIntent(
                this, R.xml.preferences_geo_objects, R.string.title_geoObjectsPreferences
        );

        startActivity(startIntent);
    }

    private class MapTapListener implements MapView.MapTapListener {
        @Override
        public void onSingleTap(MapCoordinates mapCoordinates) {
            Drawable markerDrawable = ContextCompat.getDrawable(
                    GeoObjectsActivity.this, android.R.drawable.ic_menu_mylocation);
            GeoMarker marker = new GeoMarker(mapCoordinates, markerDrawable, new PointF(0.5f, 0.5f));
            marker.setClickable(true);
            mapView.addGeoObject(marker);
        }

        @Override
        public void onDoubleTap(MapCoordinates mapCoordinates) {
            geoPolyline.addVertex(mapCoordinates);
            mapView.addGeoObject(geoPolyline);
        }

        @Override
        public void onLongTap(MapCoordinates mapCoordinates) {
            geoPolygon.addVertex(mapCoordinates);
            mapView.addGeoObject(geoPolygon);
        }
    }

    private class GeoObjectListener implements MapView.GeoObjectListener {
        @Override
        public void onGeoObjectClick(GeoObject geoObject) {
            if (!isEnabled()) {
                return;
            }

            if (geoObject instanceof GeoMarker) {
                onGeoMarkerClick((GeoMarker) geoObject);
            }

            Toast.makeText(
                    GeoObjectsActivity.this,
                    R.string.geoObjects_geoObjectClicked,
                    Toast.LENGTH_SHORT
            ).show();
        }

        @Override
        public void onGeoObjectPress(GeoObject geoObject) {
        }

        @Override
        public void onGeoObjectRelease(GeoObject geoObject) {
        }

        @Override
        public void onGeoObjectLongPress(GeoObject geoObject) {
        }

        private boolean isEnabled() {
            return PreferenceUtils.getBoolean(
                    GeoObjectsActivity.this,
                    R.string.key_geoObjects_clickable,
                    R.bool.defaultValue_geoObjects_clickable
            );
        }

        private void onGeoMarkerClick(GeoMarker geoMarker) {
            Drawable newMarkerDrawable = ContextCompat.getDrawable(
                    GeoObjectsActivity.this, android.R.drawable.star_big_on);

            geoMarker.setImage(newMarkerDrawable, new PointF(0.5f, 0.5f));

            MapRectangle mapRectangle =
                    new MapRectangle(geoMarker.getCoordinates(), geoMarker.getCoordinates());

            MapView.CameraController cameraController = mapView.getCameraController();

            CameraPosition newCameraPosition =
                    cameraController.getCameraPositionForMapRectangle(mapRectangle, 0, 0);

            cameraController.animateCamera(newCameraPosition);
        }
    }

    // region Preferences
    private static final List<Integer> PREFERENCE_KEY_IDS_TO_LISTEN = Arrays.asList(
            R.string.key_geoObjects_polylineColor,
            R.string.key_geoObjects_polylineWidth,
            R.string.key_geoObjects_polygonFillColor,
            R.string.key_geoObjects_polygonStrokeColor,
            R.string.key_geoObjects_polygonStrokeWidth,
            R.string.key_geoObjects_clickable
    );

    private class PreferenceChangeCallback implements PreferenceChangeListener.Callback {
        @Override
        public void onPreferenceChanged(@StringRes int keyId) {
            switch (keyId) {
                case R.string.key_geoObjects_polylineColor:
                    applyPolylineColorPreference();
                    break;
                case R.string.key_geoObjects_polylineWidth:
                    applyPolylineWidthPreference();
                    break;
                case R.string.key_geoObjects_polygonFillColor:
                    applyPolygonFillColorPreference();
                    break;
                case R.string.key_geoObjects_polygonStrokeColor:
                    applyPolygonStrokeColorPreference();
                    break;
                case R.string.key_geoObjects_polygonStrokeWidth:
                    applyPolygonStrokeWidthPreference();
                    break;
                case R.string.key_geoObjects_clickable:
                    // nothing :)
                    break;
            }
        }
    }

    private void applyAllGeoObjectPreferences() {
        applyPolylineColorPreference();
        applyPolylineWidthPreference();
        applyPolygonFillColorPreference();
        applyPolygonStrokeColorPreference();
        applyPolygonStrokeWidthPreference();
    }

    private void applyPolylineColorPreference() {
        String value = PreferenceUtils.getString(
                this,
                R.string.key_geoObjects_polylineColor,
                R.string.defaultValue_geoObjects_polylineColor
        );

        geoPolyline.setColor(Color.parseColor(value));
    }

    private void applyPolylineWidthPreference() {
        String value = PreferenceUtils.getString(
                this,
                R.string.key_geoObjects_polylineWidth,
                R.string.defaultValue_geoObjects_polylineWidth
        );

        geoPolyline.setWidth(Float.parseFloat(value));
    }

    private void applyPolygonFillColorPreference() {
        String value = PreferenceUtils.getString(
                this,
                R.string.key_geoObjects_polygonFillColor,
                R.string.defaultValue_geoObjects_polygonFillColor
        );

        geoPolygon.setFillColor(Color.parseColor(value));
    }

    private void applyPolygonStrokeColorPreference() {
        String value = PreferenceUtils.getString(
                this,
                R.string.key_geoObjects_polygonStrokeColor,
                R.string.defaultValue_geoObjects_polygonStrokeColor
        );

        geoPolygon.setStrokeColor(Color.parseColor(value));
    }

    private void applyPolygonStrokeWidthPreference() {
        String value = PreferenceUtils.getString(
                this,
                R.string.key_geoObjects_polygonStrokeWidth,
                R.string.defaultValue_geoObjects_polygonStrokeWidth
        );

        geoPolygon.setStrokeWidth(Float.parseFloat(value));
    }
    // endregion
}
