package com.navmii.sdksampleapp.map;

import android.content.Intent;
import android.content.SharedPreferences;
import android.os.Bundle;
import android.preference.PreferenceManager;
import android.text.TextUtils;
import android.view.MenuItem;
import android.view.View;
import android.view.WindowManager;

import androidx.annotation.StringRes;
import androidx.appcompat.app.AppCompatActivity;

import com.navmii.sdk.Sdk;
import com.navmii.sdk.common.MapCoordinates;
import com.navmii.sdk.map.CameraMovementReason;
import com.navmii.sdk.map.CameraPosition;
import com.navmii.sdk.map.MapView;
import com.navmii.sdksampleapp.PreferenceChangeListener;
import com.navmii.sdksampleapp.PreferenceUtils;
import com.navmii.sdksampleapp.R;
import com.navmii.sdksampleapp.sdk.SdkFragment;

import java.util.Arrays;
import java.util.List;


public class MapActivity extends AppCompatActivity implements View.OnClickListener {

    private static final float DEFAULT_TILT_COEF = 15f;
    private static final float DEFAULT_ZOOM_COEF = 0.5f;
    private static final float DEFAULT_HEADING_COEF = 15f;

    private static final float ANIMATION_DURATION = 0.25f;

    private final Sdk sdk = Sdk.getInstance();
    private final Sdk.StateChangeListener stateChangeListener = new SdkStateListener();

    private MapView mapView;
    private View mapControls;
    private View initializationProgressBar;
    private MapView.CameraController cameraController;

    private final MapView.CameraMovementListener cameraMovementListener = new CameraMovementListener();

    private SharedPreferences.OnSharedPreferenceChangeListener preferenceChangeListener;
    private View followPositionButton;
    private View followDirectionButton;

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.activity_map);
        getWindow().addFlags(WindowManager.LayoutParams.FLAG_KEEP_SCREEN_ON);

        if (getSupportActionBar() != null) {
            getSupportActionBar().setDisplayHomeAsUpEnabled(true);
            getSupportActionBar().setDisplayShowHomeEnabled(true);
        }

        preferenceChangeListener = PreferenceChangeListener.create(
                this,
                PREFERENCE_KEY_IDS_TO_LISTEN,
                new PreferenceChangeListenerCallback()
        );

        if (getFragmentManager().findFragmentByTag(SdkFragment.SDK_FRAGMENT_TAG) == null) {
            getFragmentManager()
                    .beginTransaction()
                    .add(new SdkFragment(), SdkFragment.SDK_FRAGMENT_TAG)
                    .commit();
        }

        initializationProgressBar = findViewById(R.id.initialization_progressbar);
        mapView = findViewById(R.id.map_view);
        cameraController = mapView.getCameraController();

        mapControls = findViewById(R.id.map_controls);
        mapControls.setVisibility(View.GONE);

        followPositionButton = findViewById(R.id.follow_position_button);
        followPositionButton.setOnClickListener(this);

        followDirectionButton = findViewById(R.id.follow_direction_button);
        followDirectionButton.setOnClickListener(this);

        findViewById(R.id.zoom_in_button).setOnClickListener(this);
        findViewById(R.id.zoom_out_button).setOnClickListener(this);
        findViewById(R.id.tilt_in_button).setOnClickListener(this);
        findViewById(R.id.tilt_out_button).setOnClickListener(this);
        findViewById(R.id.rotate_left_button).setOnClickListener(this);
        findViewById(R.id.rotate_right_button).setOnClickListener(this);
        findViewById(R.id.settings_button).setOnClickListener(this);
    }

    @Override
    protected void onStart() {
        super.onStart();

        if (sdk.getState() == Sdk.State.INITIALIZED) {
            sdk.setActiveMapView(mapView);
            applyAllMapViewPreferences();
        }

        sdk.addStateChangeListener(stateChangeListener);

        updateControlsVisibility();

        cameraController.addCameraMovementListener(cameraMovementListener);

        PreferenceManager.getDefaultSharedPreferences(this)
                .registerOnSharedPreferenceChangeListener(preferenceChangeListener);
    }

    @Override
    protected void onStop() {
        super.onStop();

        if (sdk.getState() == Sdk.State.INITIALIZED) {
            sdk.setActiveMapView(null);
        }

        sdk.removeStateChangeListener(stateChangeListener);

        cameraController.removeCameraMovementListener(cameraMovementListener);

        PreferenceManager.getDefaultSharedPreferences(this)
                .unregisterOnSharedPreferenceChangeListener(preferenceChangeListener);
    }

    @Override
    public boolean onOptionsItemSelected(MenuItem item) {
        switch (item.getItemId()) {
            case android.R.id.home:
                finish();
                return true;
        }

        return super.onOptionsItemSelected(item);
    }

    private class SdkStateListener implements Sdk.StateChangeListener {
        @Override
        public void onStateChanged(Sdk.State state) {
            if (state == Sdk.State.INITIALIZED) {
                sdk.setActiveMapView(mapView);
                applyAllMapViewPreferences();
            }

            updateControlsVisibility();
        }
    }

    private void updateControlsVisibility() {
        initializationProgressBar.setVisibility(
                sdk.getState() == Sdk.State.INITIALIZING ? View.VISIBLE : View.GONE);

        mapControls.setVisibility(
                sdk.getState() == Sdk.State.INITIALIZED ? View.VISIBLE : View.GONE);

        updateFollowPositionButton();
        updateFollowDirectionButton();
    }

    private void updateFollowPositionButton() {
        followPositionButton.setActivated(
                sdk.getState() == Sdk.State.INITIALIZED && mapView.isFollowGpsPosition());
    }

    private void updateFollowDirectionButton() {
        followDirectionButton.setActivated(
                sdk.getState() == Sdk.State.INITIALIZED && mapView.isFollowGpsDirection());
    }

    private void moveCameraToGpsPosition() {

        if (sdk.getState() != Sdk.State.INITIALIZED)
            return;

        MapCoordinates extrapolatedPosition = sdk.getPositionManager().getExtrapolatedPosition();

        if (extrapolatedPosition == null) {
            return;
        }

        CameraPosition cameraPosition =
                new CameraPosition.Builder()
                        .setTargetLocation(extrapolatedPosition)
                        .build();

        cameraController.animateCamera(cameraPosition, ANIMATION_DURATION, null);
    }

    private float getCurrentZoomLevel() {
        return cameraController.getCameraPosition().getZoom();
    }

    private void setZoomLevel(float value) {
        CameraPosition cameraPosition =
                new CameraPosition.Builder()
                        .setZoom(value)
                        .build();

        cameraController.animateCamera(cameraPosition, ANIMATION_DURATION, null);
    }

    private class CameraMovementListener implements MapView.CameraMovementListener {
        @Override
        public void onCameraMoved(CameraMovementReason cameraMovementReason) {
            updateFollowPositionButton();
            updateFollowDirectionButton();
        }
    }

    // region Button clicks
    @Override
    public void onClick(View view) {
        switch (view.getId()) {
            case R.id.zoom_in_button:
                zoomInClicked();
                break;
            case R.id.zoom_out_button:
                zoomOutClicked();
                break;
            case R.id.tilt_in_button:
                tiltInClicked();
                break;
            case R.id.tilt_out_button:
                tiltOutClicked();
                break;
            case R.id.rotate_left_button:
                rotateLeftClicked();
                break;
            case R.id.rotate_right_button:
                rotateRightClicked();
                break;
            case R.id.follow_position_button:
                followPositionClicked();
                break;
            case R.id.follow_direction_button:
                followDirectionClicked();
                break;
            case R.id.settings_button:
                startMapPreferenceActivity();
                break;
        }
    }

    private void zoomInClicked() {
        float currentZoomLevel = getCurrentZoomLevel();
        float zoomLevel = currentZoomLevel + DEFAULT_ZOOM_COEF;
        setZoomLevel(zoomLevel);
    }

    private void zoomOutClicked() {
        float currentZoomLevel = getCurrentZoomLevel();
        float zoomLevel = currentZoomLevel - DEFAULT_ZOOM_COEF;
        setZoomLevel(zoomLevel);
    }

    private void tiltInClicked() {
        setTiltLevel(cameraController.getCameraPosition().getTilt() - DEFAULT_TILT_COEF);
    }

    private void tiltOutClicked() {
        setTiltLevel(cameraController.getCameraPosition().getTilt() + DEFAULT_TILT_COEF);
    }

    private void setTiltLevel(float tiltLevel) {
        CameraPosition cameraPosition =
                new CameraPosition.Builder()
                        .setTilt(tiltLevel)
                        .build();

        cameraController.animateCamera(cameraPosition, ANIMATION_DURATION, null);
    }

    private void rotateLeftClicked() {
        setRotation(cameraController.getCameraPosition().getHeading() - DEFAULT_HEADING_COEF);
    }

    private void rotateRightClicked() {
        setRotation(cameraController.getCameraPosition().getHeading() + DEFAULT_HEADING_COEF);
    }

    private void setRotation(float rotation) {
        CameraPosition cameraPosition =
                new CameraPosition.Builder()
                        .setHeading(rotation)
                        .build();

        cameraController.animateCamera(cameraPosition, ANIMATION_DURATION, null);
    }

    private void followPositionClicked() {
        mapView.setFollowGpsPosition(!mapView.isFollowGpsPosition());
        updateFollowPositionButton();
        moveCameraToGpsPosition();
    }

    private void followDirectionClicked() {
        mapView.setFollowGpsDirection(!mapView.isFollowGpsDirection());
        updateFollowDirectionButton();
    }

    private void startMapPreferenceActivity() {
        final Intent startIntent = MapPreferenceActivity.createStartIntent(
                this, mapView.getAvailableMapSchemes());

        startActivity(startIntent);
    }
    // endregion

    // region Preferences
    private static final List<Integer> PREFERENCE_KEY_IDS_TO_LISTEN = Arrays.asList(
            R.string.key_map_panningGesture,
            R.string.key_map_rotationGesture,
            R.string.key_map_zoomGesture,
            R.string.key_map_tiltGesture,
            R.string.key_map_showGpsPosition,
            R.string.key_map_scheme
    );

    private class PreferenceChangeListenerCallback implements PreferenceChangeListener.Callback {
        @Override
        public void onPreferenceChanged(@StringRes int keyId) {
            switch (keyId) {
                case R.string.key_map_panningGesture:
                    applyPanningGesturePreference();
                    break;
                case R.string.key_map_rotationGesture:
                    applyRotationGesturePreference();
                    break;
                case R.string.key_map_zoomGesture:
                    applyZoomGesturePreference();
                    break;
                case R.string.key_map_tiltGesture:
                    applyTiltGesturePreference();
                    break;
                case R.string.key_map_showGpsPosition:
                    applyShowGpsPositionPreference();
                    break;
                case R.string.key_map_scheme:
                    applyMapSchemePreference();
                    break;
            }
        }
    }

    private void applyAllMapViewPreferences() {
        applyPanningGesturePreference();
        applyRotationGesturePreference();
        applyZoomGesturePreference();
        applyTiltGesturePreference();
        applyShowGpsPositionPreference();
        applyMapSchemePreference();
    }

    private void applyPanningGesturePreference() {
        final boolean value = PreferenceUtils.getBoolean(
                this,
                R.string.key_map_panningGesture,
                R.bool.defaultValue_map_panningGesture
        );

        mapView.setMapPanningGestureEnabled(value);
    }

    private void applyRotationGesturePreference() {
        final boolean value = PreferenceUtils.getBoolean(
                this,
                R.string.key_map_rotationGesture,
                R.bool.defaultValue_map_rotationGesture
        );

        mapView.setMapRotationGestureEnabled(value);
    }

    private void applyZoomGesturePreference() {
        final boolean value = PreferenceUtils.getBoolean(
                this,
                R.string.key_map_zoomGesture,
                R.bool.defaultValue_map_zoomGesture
        );

        mapView.setMapZoomGestureEnabled(value);
    }

    private void applyTiltGesturePreference() {
        final boolean value = PreferenceUtils.getBoolean(
                this,
                R.string.key_map_tiltGesture,
                R.bool.defaultValue_map_tiltGesture
        );

        mapView.setMapTiltGestureEnabled(value);
    }

    private void applyShowGpsPositionPreference() {
        final boolean value = PreferenceUtils.getBoolean(
                this,
                R.string.key_map_showGpsPosition,
                R.bool.defaultValue_map_showGpsPosition
        );

        mapView.setGpsPositionShown(value);
    }

    private void applyMapSchemePreference() {
        SharedPreferences preferences = PreferenceManager.getDefaultSharedPreferences(this);

        String keyMapScheme = getString(R.string.key_map_scheme);

        String currentMapScheme = preferences.getString(keyMapScheme, "");

        if (TextUtils.isEmpty(currentMapScheme)) {
            List<String> availableMapSchemes = mapView.getAvailableMapSchemes();

            if (!availableMapSchemes.isEmpty()) {
                currentMapScheme = availableMapSchemes.get(0);

                SharedPreferences.Editor editor = preferences.edit();
                editor.putString(keyMapScheme, currentMapScheme);
                editor.apply();

                mapView.loadMapScheme(currentMapScheme);
            }
        } else {
            mapView.loadMapScheme(currentMapScheme);
        }
    }
    // endregion
}
