package com.navmii.sdksampleapp.navigation;

import android.app.ProgressDialog;
import android.content.Context;
import android.content.DialogInterface;
import android.content.Intent;
import android.content.SharedPreferences;
import android.graphics.PointF;
import android.graphics.Rect;
import android.location.Location;
import android.os.Bundle;
import android.preference.PreferenceManager;
import android.text.TextUtils;
import android.util.Log;
import android.view.Menu;
import android.view.MenuInflater;
import android.view.MenuItem;
import android.view.View;
import android.view.ViewGroup;
import android.view.WindowManager;
import android.widget.ImageButton;
import android.widget.ImageView;
import android.widget.Toast;

import androidx.annotation.DrawableRes;
import androidx.annotation.Nullable;
import androidx.annotation.StringRes;
import androidx.appcompat.app.AlertDialog;
import androidx.appcompat.app.AppCompatActivity;
import androidx.core.content.ContextCompat;
import androidx.core.content.res.ResourcesCompat;
import androidx.viewpager.widget.ViewPager;

import com.google.android.material.snackbar.Snackbar;
import com.navmii.sdk.Sdk;
import com.navmii.sdk.common.Color;
import com.navmii.sdk.common.MapCoordinates;
import com.navmii.sdk.common.MapRectangle;
import com.navmii.sdk.map.AutoZoomMode;
import com.navmii.sdk.map.CameraMovementReason;
import com.navmii.sdk.map.CameraPosition;
import com.navmii.sdk.map.GeoMarker;
import com.navmii.sdk.map.MapView;
import com.navmii.sdk.navigation.NavigationManager;
import com.navmii.sdk.positioning.GpsStatus;
import com.navmii.sdk.positioning.Position;
import com.navmii.sdk.positioning.PositionManager;
import com.navmii.sdk.routecalculation.ExternalRoute;
import com.navmii.sdk.routecalculation.ExternalRoutePoint;
import com.navmii.sdk.routecalculation.Route;
import com.navmii.sdk.routecalculation.RouteCalculationError;
import com.navmii.sdk.routecalculation.RouteCalculationFailureInfo;
import com.navmii.sdk.routecalculation.RouteCalculationOptions;
import com.navmii.sdk.routecalculation.RouteCalculationResult;
import com.navmii.sdk.routecalculation.RouteCalculationSession;
import com.navmii.sdk.routecalculation.RoutePlanPoint;
import com.navmii.sdk.routecalculation.RoutingManager;
import com.navmii.sdk.routecalculation.RoutingOptimization;
import com.navmii.sdk.routecalculation.SnappingOptions;
import com.navmii.sdk.routecalculation.VehicleType;
import com.navmii.sdk.routegeofence.RouteGeofenceManager;
import com.navmii.sdk.routenavigation.GuidanceOutput;
import com.navmii.sdk.routenavigation.RouteNavigator;
import com.navmii.sdk.routevisualization.RouteDisplayParameters;
import com.navmii.sdk.routevisualization.RouteDisplayParametersBuilder;
import com.navmii.sdk.routevisualization.RouteVisualizer;
import com.navmii.sdk.routevisualization.RouteVisualizerRequest;
import com.navmii.sdksampleapp.PreferenceChangeListener;
import com.navmii.sdksampleapp.PreferenceUtils;
import com.navmii.sdksampleapp.R;
import com.navmii.sdksampleapp.Utils;
import com.navmii.sdksampleapp.routing.RoutingInfoFragment;
import com.navmii.sdksampleapp.routing.RoutingPagerAdapter;
import com.navmii.sdksampleapp.sdk.SdkFragment;
import com.navmii.sdksampleapp.views.NavigationInfoController;
import com.navmii.sdksampleapp.views.SpeedLimitController;
import com.viewpagerindicator.CirclePageIndicator;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;


public class NavigationActivity extends AppCompatActivity
        implements View.OnClickListener, MapView.MapTapListener, MapView.CameraMovementListener {

    private static final float TILT_NAVIGATION = 60f;
    private static final float ANIMATION_DURATION = 0.5f;
    private static final float ZOOM_LEVEL = 11;
    private static final int ROUTE_OVERVIEW_PADDING_DP = 50;
    private static final List<Integer> PREFERENCE_KEY_IDS_TO_LISTEN = Arrays.asList(
            R.string.key_navigation_guidanceOutput,
            R.string.key_navigation_language,
            R.string.key_navigation_autozoomEnabled,
            R.string.key_navigation_routeGeofenceEnabled,
            R.string.key_navigation_routeGeofenceAccuracy
    );

    private final Sdk sdk = Sdk.getInstance();
    private final Sdk.StateChangeListener stateChangeListener = new SdkStateListener();
    private final SpeedLimitController speedLimitController = new SpeedLimitController();
    private final NavigationListener navigationListener = new NavigationListener();
    private final PositionListener positionListener = new PositionListener();
    private final RouteVisualizerRequestListener routeVisualizerRequestListener =
            new RouteVisualizerRequestListener();
    private final RouteGeofenceListener routeGeofenceListener = new RouteGeofenceListener();

    private SharedPreferences.OnSharedPreferenceChangeListener preferenceChangeListener;
    private MapView mapView;
    private View clearButton;
    private ImageButton demoRouteButton;
    private ImageButton snapRouteButton;
    private ImageButton snapToPositionButton;
    private ProgressDialog routeCalculationDialog;
    private View initializationProgressBar;
    private View routesInfo;
    private ViewPager routeInfoPager;
    private MenuItem applyRouteMenuItem;
    private View navigationInfoView;
    private ImageView gpsStatusView;
    private RoutingPagerAdapter routingPagerAdapter;
    private NavigationInfoController navigationInfoController;
    private boolean canApplyRoute;
    private boolean isRouteApplied;
    private List<Route> routes = new ArrayList<>();
    private int selectedRouteIndex = -1;
    private RouteVisualizerRequest routeVisualizerRequest;

    private RouteCalculationSession routeCalculationSession;
    private RouteCalculationResult routeCalculationResult;

    private final List<GeoMarker> intermediateWaypoints = new ArrayList<>();

    @Nullable
    public static GuidanceOutput getGuidanceOutputFromPersistedString(
            Context context, String value) {

        if (TextUtils.equals(value,
                context.getString(R.string.entryValue_navigation_guidanceOutput_none))) {
            return GuidanceOutput.NONE;
        }

        if (TextUtils.equals(value,
                context.getString(R.string.entryValue_navigation_guidanceOutput_text))) {
            return GuidanceOutput.TEXT;
        }

        return null;
    }

    public static AutoZoomMode getAutozoomModeFromPersistedString(
            Context context, String value) {

        if (TextUtils.equals(value,
                context.getString(R.string.entryValue_navigation_autozoomMode_atDirections))) {
            return AutoZoomMode.DIRECTION;
        }

        if (TextUtils.equals(value,
                context.getString(R.string.entryValue_navigation_autozoomMode_onSpeed))) {
            return AutoZoomMode.SPEED;
        }

        return AutoZoomMode.SPEED;
    }

    @Nullable
    private static RoutingOptimization getRouteOptimizationByPreferenceString(
            Context context, String value) {

        if (TextUtils.equals(value,
                context.getString(R.string.entryValue_routing_optimization_fastest))) {
            return RoutingOptimization.FASTEST;
        }

        if (TextUtils.equals(value,
                context.getString(R.string.entryValue_routing_optimization_easy))) {
            return RoutingOptimization.EASIEST;
        }

        if (TextUtils.equals(value,
                context.getString(R.string.entryValue_routing_optimization_economical))) {
            return RoutingOptimization.ECONOMICAL;
        }

        if (TextUtils.equals(value,
                context.getString(R.string.entryValue_routing_optimization_shortest))) {
            return RoutingOptimization.SHORTEST;
        }

        return null;
    }

    @Nullable
    private static VehicleType getVehicleTypeByPreferenceString(
            Context context, String value) {

        if (TextUtils.equals(value,
                context.getString(R.string.entryValue_routing_vehicleType_car))) {
            return VehicleType.CAR;
        }

        if (TextUtils.equals(value,
                context.getString(R.string.entryValue_routing_vehicleType_motorcycle))) {
            return VehicleType.MOTORCYCLE;
        }

        if (TextUtils.equals(value,
                context.getString(R.string.entryValue_routing_vehicleType_cycle))) {
            return VehicleType.CYCLE;
        }

        if (TextUtils.equals(value,
                context.getString(R.string.entryValue_routing_vehicleType_rvLorry))) {
            return VehicleType.LORRY;
        }

        if (TextUtils.equals(value,
                context.getString(R.string.entryValue_routing_vehicleType_truck))) {
            return VehicleType.TRUCK;
        }

        if (TextUtils.equals(value,
                context.getString(R.string.entryValue_routing_vehicleType_deliveryTruck))) {
            return VehicleType.DELIVERY_TRUCK;
        }

        return null;
    }

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.activity_navigation);
        getWindow().addFlags(WindowManager.LayoutParams.FLAG_KEEP_SCREEN_ON);

        if (getSupportActionBar() != null) {
            getSupportActionBar().setDisplayHomeAsUpEnabled(true);
            getSupportActionBar().setDisplayShowHomeEnabled(true);
        }

        preferenceChangeListener = PreferenceChangeListener.create(
                this,
                PREFERENCE_KEY_IDS_TO_LISTEN,
                new PreferenceChangeCallback()
        );

        if (getFragmentManager().findFragmentByTag(SdkFragment.SDK_FRAGMENT_TAG) == null) {
            getFragmentManager()
                    .beginTransaction()
                    .add(new SdkFragment(), SdkFragment.SDK_FRAGMENT_TAG)
                    .commit();
        }

        mapView = findViewById(R.id.map_view);
        mapView.setMapZoomOnDoubleTapEnabled(false);
        initializationProgressBar = findViewById(R.id.initialization_progressbar);

        clearButton = findViewById(R.id.clear_button);
        clearButton.setOnClickListener(this);

        demoRouteButton = findViewById(R.id.demo_route_button);
        demoRouteButton.setOnClickListener(this);

        snapRouteButton = findViewById(R.id.snap_route_button);
        snapRouteButton.setOnClickListener(this);

        snapToPositionButton = findViewById(R.id.snap_to_position_button);
        snapToPositionButton.setOnClickListener(this);

        gpsStatusView = findViewById(R.id.gps_status);

        findViewById(R.id.settings_button).setOnClickListener(this);
        ViewGroup speedLimitContainer = findViewById(R.id.speed_limit_container);
        speedLimitController.setViewContainer(speedLimitContainer);

        navigationInfoView = findViewById(R.id.navigation_info);

        routesInfo = findViewById(R.id.routes_info);

        routingPagerAdapter = new RoutingPagerAdapter(getSupportFragmentManager());

        routeInfoPager = findViewById(R.id.route_info_pager);
        routeInfoPager.setAdapter(routingPagerAdapter);
        routeInfoPager.addOnPageChangeListener(new ViewPager.OnPageChangeListener() {
            @Override
            public void onPageScrolled(int position, float positionOffset, int positionOffsetPixels) {
            }

            @Override
            public void onPageSelected(int position) {
                if (sdk.getState() == Sdk.State.INITIALIZED && isRouteCalculated()) {
                    selectRoute(position);
                }
            }

            @Override
            public void onPageScrollStateChanged(int state) {
            }
        });

        CirclePageIndicator routeInfoPagerIndicator =
                findViewById(R.id.route_info_pager_indicator);
        routeInfoPagerIndicator.setViewPager(routeInfoPager);

        mapView.postDelayed(new Runnable() {
            @Override
            public void run() {
                updateRouteOverview();
            }
        }, 1000);

        updateRoutingInfoView();
        updateNavigationInfoView();
        updateRouteButtons();
        updateGpsStatusView();
    }

    @Override
    protected void onDestroy() {
        super.onDestroy();

        speedLimitController.setViewContainer(null);
    }

    @Override
    protected void onStart() {
        super.onStart();

        if (sdk.getState() == Sdk.State.INITIALIZED) {
            sdk.setActiveMapView(mapView);

            addSdkListeners();

            applyAllNavigationPreferences();

            speedLimitController.setNavigationManager(sdk.getNavigationManager());
        }

        updateControlsVisibility();

        sdk.addStateChangeListener(stateChangeListener);

        mapView.addMapTapListener(this);
        mapView.getCameraController().addCameraMovementListener(this);

        PreferenceManager.getDefaultSharedPreferences(this)
                .registerOnSharedPreferenceChangeListener(preferenceChangeListener);
    }

    @Override
    protected void onStop() {
        super.onStop();

        if (sdk.getState() == Sdk.State.INITIALIZED) {
            sdk.setActiveMapView(null);

            removeSdkListeners();

            speedLimitController.setNavigationManager(null);
        }

        sdk.removeStateChangeListener(stateChangeListener);

        mapView.removeMapTapListener(this);

        PreferenceManager.getDefaultSharedPreferences(this)
                .unregisterOnSharedPreferenceChangeListener(preferenceChangeListener);
    }

    @Override
    public boolean onCreateOptionsMenu(Menu menu) {
        MenuInflater inflater = getMenuInflater();
        inflater.inflate(R.menu.menu_navigation, menu);

        applyRouteMenuItem = menu.findItem(R.id.apply_route);
        updateMenu();

        return true;
    }

    @Override
    public boolean onOptionsItemSelected(MenuItem item) {
        switch (item.getItemId()) {
            case android.R.id.home:
                finish();
                return true;
            case R.id.apply_route:
                applySelectedRoute();
                return true;
        }

        return super.onOptionsItemSelected(item);
    }

    private void updateMenu() {
        if (applyRouteMenuItem != null) {
            applyRouteMenuItem.setVisible(canApplyRoute);
        }
    }

    private void updateControlsVisibility() {
        initializationProgressBar.setVisibility(
                sdk.getState() == Sdk.State.INITIALIZING ? View.VISIBLE : View.GONE);

        boolean isSdkInitialized = sdk.getState() == Sdk.State.INITIALIZED;

        snapRouteButton.setVisibility(
                isSdkInitialized ? View.VISIBLE : View.GONE);

        snapToPositionButton.setVisibility(
                isSdkInitialized ? View.VISIBLE : View.GONE);

        findViewById(R.id.settings_button).setVisibility(
                isSdkInitialized ? View.VISIBLE : View.GONE);

        if (gpsStatusView != null) {
            gpsStatusView.setVisibility(isSdkInitialized ? View.VISIBLE : View.GONE);
        }
    }

    @Override
    public void onClick(View view) {
        switch (view.getId()) {
            case R.id.snap_to_position_button:
                onSnapToPositionButtonClicked();
                break;
            case R.id.snap_route_button:
                onSnapRouteButtonClicked();
                break;
            case R.id.settings_button:
                startNavigationPreferenceActivity();
                break;
            case R.id.clear_button:
                clearRoutes();
                break;
            case R.id.demo_route_button:
                toggleDemoRoute();
                break;
            case R.id.apply_route:
                applySelectedRoute();
                break;
        }
    }

    private void onSnapToPositionButtonClicked() {
        mapView.setFollowGpsPosition(true);
    }

    private void onSnapRouteButtonClicked() {
        final ArrayList<ExternalRoutePoint> externalRoutePoints = new ArrayList<>();

        externalRoutePoints.add(
                new ExternalRoutePoint(new MapCoordinates(51.500047, -0.104653), 0, false));
        externalRoutePoints.add(
                new ExternalRoutePoint(new MapCoordinates(51.500269, -0.104512), 10, false));
        externalRoutePoints.add(
                new ExternalRoutePoint(new MapCoordinates(51.502253, -0.104561), 20, false));
        externalRoutePoints.add(
                new ExternalRoutePoint(new MapCoordinates(51.503679, -0.104540), 30, true));
        externalRoutePoints.add(
                new ExternalRoutePoint(new MapCoordinates(51.503627, -0.103485), 40, false));
        externalRoutePoints.add(
                new ExternalRoutePoint(new MapCoordinates(51.503635, -0.102517), 50, false));
        externalRoutePoints.add(
                new ExternalRoutePoint(new MapCoordinates(51.503351, -0.102580), 60, false));

        final ExternalRoute externalRoute = new ExternalRoute(
                /*durationInSeconds*/60,
                /*durationIgnoringTrafficDelayInSeconds*/50,
                /*mayContainSelfIntersections*/false,
                externalRoutePoints);

        final List<CharSequence> dialogItems = new ArrayList<>();
        dialogItems.add(getString(R.string.navigation_snap_route));
        dialogItems.add(getString(R.string.navigation_snap_route_from_position));

        AlertDialog alertDialog = new AlertDialog.Builder(this)
                .setItems(dialogItems.toArray(new CharSequence[0]), (dialog, which) -> {
                    switch (which) {
                        case 0:
                            routeCalculationSession = sdk.getRoutingManager().snapRoute(
                                    externalRoute,
                                    getSnappingOptions(),
                                    new RouteCalculationListener());
                            onRouteCalculationStarted();
                            break;
                        case 1:
                            ExternalRoutePoint startingPoint = new ExternalRoutePoint(
                                    new MapCoordinates(51.498534, -0.107945), 0, false);
                            routeCalculationSession = sdk.getRoutingManager().snapRouteFromPosition(
                                    startingPoint,
                                    externalRoute,
                                    getSnappingOptions(),
                                    new RouteCalculationListener());
                            onRouteCalculationStarted();
                            break;
                    }
                })
                .create();

        alertDialog.show();
    }

    private SnappingOptions getSnappingOptions() {
        SnappingOptions snappingOptions = SnappingOptions.create();
        snappingOptions.setIsWalkingMode(false);
        return snappingOptions;
    }

    private void startNavigationPreferenceActivity() {
        final Intent startIntent = NavigationPreferenceActivity.createStartIntent(this);
        startActivity(startIntent);
    }

    @Override
    public void onSingleTap(MapCoordinates mapCoordinates) {
        PositionManager positionManager = sdk.getPositionManager();

        if (positionManager == null) {
            return;
        }

        final Position currentPosition = positionManager.getCurrentPosition();

        if (currentPosition == null) {
            Toast.makeText(this, R.string.no_current_position, Toast.LENGTH_SHORT).show();
            return;
        }

        final ArrayList<RoutePlanPoint> waypoints = new ArrayList<>();
        waypoints.add(RoutePlanPoint.createWithPosition(currentPosition));
        for (GeoMarker waypoint: this.intermediateWaypoints) {
            waypoints.add(RoutePlanPoint.createWithCoordinates(
                    waypoint.getCoordinates()));
        }
        waypoints.add(RoutePlanPoint.createWithCoordinates(mapCoordinates));

        for (RoutePlanPoint waypoint: waypoints) {
            Log.i("NavigationActivity", "Route waypoint: " + waypoint.coordinates());
        }

        routeCalculationSession = sdk.getRoutingManager().calculateRoute(
                waypoints,
                buildRouteCalculationOptions(),
                new RouteCalculationListener());

        onRouteCalculationStarted();
    }

    @Override
    public void onDoubleTap(MapCoordinates mapCoordinates) {
        GeoMarker marker = new GeoMarker(
                mapCoordinates,
                ResourcesCompat.getDrawable(getResources(), R.drawable.ic_plus, null),
                new PointF(0.5f, 0.5f));
        intermediateWaypoints.add(marker);
        mapView.addGeoObject(marker);
    }

    @Override
    public void onLongTap(MapCoordinates mapCoordinates) {

    }

    private void closeRouteCalculationDialog() {
        if (routeCalculationDialog != null) {
            routeCalculationDialog.dismiss();
        }
    }

    private void updateRouteButtons() {
        boolean routeButtonsVisible =
                sdk != null && isRouteCalculated();

        int visibility = routeButtonsVisible ? View.VISIBLE : View.GONE;

        if (clearButton != null) {
            clearButton.setVisibility(visibility);
        }

        if (demoRouteButton != null) {
            demoRouteButton.setVisibility(visibility);

            if (routeButtonsVisible) {
                @DrawableRes int resId =
                        sdk.getNavigationManager() != null &&
                                sdk.getNavigationManager().isDemoRouteActive()
                                ? R.drawable.ic_demo_stop
                                : R.drawable.ic_demo_start;

                demoRouteButton.setImageDrawable(ContextCompat.getDrawable(this, resId));
            }
        }

        if (snapRouteButton != null) {
            snapRouteButton.setVisibility(routeButtonsVisible ? View.GONE : View.VISIBLE);
        }
    }

    private void updateRoutesOnMap() {
        RouteVisualizer routeVisualizer = sdk.getRouteVisualizer();

        RouteVisualizerRequest request = routeVisualizerRequest;

        if (request == null) {
            request = RouteVisualizerRequest.createRequestRemovingAll();
            routeVisualizerRequest = request;
        }

        if (isRouteCalculated() && selectedRouteIndex >= 0) {
            Route selectedRoute = routes.get(selectedRouteIndex);
            ArrayList<Route> alternativeRoutes = new ArrayList<>();

            for (int index = 0; index < routes.size(); ++index) {
                if (index != selectedRouteIndex) {
                    alternativeRoutes.add(routes.get(index));
                }
            }

            String routeColorHex = PreferenceUtils.getString(
                    this,
                    R.string.key_navigation_routeColor,
                    R.string.defaultValue_navigation_routeColor
            );

            String routeHaloColorHex = PreferenceUtils.getString(
                    this,
                    R.string.key_navigation_routeHaloColor,
                    R.string.defaultValue_navigation_routeHaloColor
            );

            boolean showTraffic = PreferenceUtils.getBoolean(
                    this,
                    R.string.key_navigation_showTrafficOnRoute,
                    R.bool.defaultValue_navigation_showTrafficOnRoute
            );

            boolean showDirectionArrows = PreferenceUtils.getBoolean(
                    this,
                    R.string.key_navigation_showDirectionArrows,
                    R.bool.defaultValue_navigation_showDirectionArrows
            );

            float minRouteWidth = PreferenceUtils.getInteger(
                    this,
                    R.string.key_navigation_minRouteWidth,
                    R.integer.defaultValue_navigation_minRouteWidth);

            float maxRouteWidth = PreferenceUtils.getInteger(
                    this,
                    R.string.key_navigation_maxRouteWidth,
                    R.integer.defaultValue_navigation_maxRouteWidth);

            RouteDisplayParametersBuilder routeDisplayParametersBuilder =
                    new RouteDisplayParametersBuilder()
                            .setShowTraffic(showTraffic)
                            .setShowDirectionArrows(showDirectionArrows)
                            .setRouteColor(new Color(routeColorHex));

            if (!TextUtils.isEmpty(routeHaloColorHex)) {
                routeDisplayParametersBuilder
                        .setShowRouteHalo(true)
                        .setRouteHaloColor(new Color(routeHaloColorHex));
            } else {
                routeDisplayParametersBuilder.setShowRouteHalo(false);
            }

            if (minRouteWidth > 0) {
                routeDisplayParametersBuilder.setMinRouteWidth(minRouteWidth);
            }
            if (maxRouteWidth > 0) {
                routeDisplayParametersBuilder.setMaxRouteWidth(maxRouteWidth);
            }

            RouteDisplayParameters altRouteDisplayParams = new RouteDisplayParametersBuilder()
                    .setRouteColor(new Color(0.9f, 0.9f, 1.0f))
                    .setMaxRouteWidth(10)
                    .build();
            request = request
                    .requestDisplayingPolylineOfRoute(
                            selectedRoute, routeDisplayParametersBuilder.build())
                    .requestDisplayingPolylinesOfAlternativeRoutes(
                            alternativeRoutes, altRouteDisplayParams);
        }

        routeVisualizer.performRequest(request);
    }

    private void updateRouteGeofence() {
        if(!routes.isEmpty()) {
            sdk.getRouteGeofenceManager().setRoute(routes.get(0));
        } else {
            sdk.getRouteGeofenceManager().clearRoute();
        }
    }

    private void updateGpsStatusView() {
        if (gpsStatusView == null || sdk == null || sdk.getPositionManager() == null) {
            return;
        }

        switch (sdk.getPositionManager().getGpsStatus()) {
            case OK:
                gpsStatusView.setImageResource(R.drawable.gps_on);
                break;
            case NO_SIGNAL:
                gpsStatusView.setImageResource(R.drawable.gps_low);
                break;
            case OFF:
                gpsStatusView.setImageResource(R.drawable.gps_off);
                break;
        }
    }

    private void clearRoutes() {
        if (sdk == null) {
            return;
        }

        sdk.getRouteVisualizer().performRequest(RouteVisualizerRequest.createRequestRemovingAll());
        routeCalculationResult = null;
        isRouteApplied = false;
        routes.clear();
        selectedRouteIndex = -1;
        routeVisualizerRequest = null;

        mapView.removeAllGeoObjects();
        intermediateWaypoints.clear();

        clearRoutesInfo();

        onRouteCleared();
    }

    private void clearRoutesInfo() {
        routingPagerAdapter.clear();
        updateRoutingInfoView();
        updateNavigationInfoView();
    }

    private void updateRoutingInfoView() {
        if (!isRouteCalculated() || isRouteApplied()) {
            canApplyRoute = false;
            routesInfo.setVisibility(View.GONE);
            updateMenu();
            return;
        }

        if (routeInfoPager.getAdapter() == null || routeInfoPager.getAdapter().getCount() == 0) {
            for (Route currentRouteInfo : routes) {
                routingPagerAdapter.addItem(
                        RoutingInfoFragment
                                .newRoutingInfoFragment(
                                        currentRouteInfo.getLength(),
                                        currentRouteInfo.getEstimatedTime(),
                                        currentRouteInfo.getName()));
            }
            routeInfoPager.setAdapter(routingPagerAdapter);
        }

        canApplyRoute = true;
        routesInfo.setVisibility(View.VISIBLE);
        updateMenu();

        routeInfoPager.post(new Runnable() {
            @Override
            public void run() {
                updateRouteOverview();
            }
        });
    }

    private void updateNavigationInfoView() {
        if (!isRouteApplied()) {
            navigationInfoController = null;
            navigationInfoView.setVisibility(View.GONE);
            return;
        }

        if (navigationInfoController == null) {
            navigationInfoController = new NavigationInfoController(navigationInfoView);
        }

        boolean updated = navigationInfoController.update(
                routes.get(0),
                sdk.getRouteNavigator());

        if (updated) {
            navigationInfoView.setVisibility(View.VISIBLE);
        }
    }

    private boolean isRouteCalculated() {
        return routeCalculationResult != null && !routes.isEmpty();
    }

    private void toggleDemoRoute() {
        if (sdk == null) {
            return;
        }

        NavigationManager navigationManager = sdk.getNavigationManager();
        RoutingManager routingManager = sdk.getRoutingManager();
        RouteNavigator routeNavigator = sdk.getRouteNavigator();

        if (navigationManager == null || routingManager == null || routeNavigator == null) {
            return;
        }

        if (navigationManager.isDemoRouteActive()) {
            navigationManager.stopDemoRoute();
            return;
        }

        if (!isRouteCalculated()) {
            showNoRouteDialog();
            return;
        }

        applySelectedRoute();

        final Route route = routeNavigator.getRoute();
        if (route != null) {
            navigationManager.startDemoRoute(route);
        }

        setCameraTilt();
    }

    private void selectRoute(int index) {
        if (index != selectedRouteIndex) {
            selectedRouteIndex = index;
            updateRoutingInfoView();
            updateRoutesOnMap();
        }
    }

    private boolean isRouteApplied() {
        return isRouteApplied;
    }

    private void applySelectedRoute() {
        if (isRouteApplied()) {
            return;
        }

        if (selectedRouteIndex >= 0) {
            sdk.getRouteNavigator().startNavigatingTheRoute(
                    routes.get(selectedRouteIndex), true);
            isRouteApplied = true;
        }

        updateRoutingInfoView();
    }

    private void updateRouteOverview() {
        if (sdk == null ||
                sdk.getRoutingManager() == null ||
                mapView.getWidth() == 0 ||
                mapView.getHeight() == 0 ||
                routeInfoPager.getHeight() == 0) {

            return;
        }

        if (selectedRouteIndex < 0) {
            return;
        }

        MapRectangle routeBounds = routes.get(selectedRouteIndex).getBoundingBox();

        final int defaultPaddingPx =
                Utils.dpToPx(ROUTE_OVERVIEW_PADDING_DP, getResources().getDisplayMetrics());

        Rect padding = new Rect(
                defaultPaddingPx,
                routeInfoPager.getHeight() + defaultPaddingPx,
                defaultPaddingPx,
                defaultPaddingPx);

        MapView.CameraController cameraController = mapView.getCameraController();

        CameraPosition routeOverviewCameraPosition =
                cameraController.getCameraPositionForMapRectangle(
                        routeBounds, 0, 0, padding, /*maxZoomLevel*/Float.MAX_VALUE);

        cameraController.animateCamera(routeOverviewCameraPosition);
    }

    private void showNoRouteDialog() {
        AlertDialog.Builder builder = new AlertDialog.Builder(this);
        builder.setTitle(R.string.navigation_dialog_error_no_route_title);
        builder.setMessage(R.string.navigation_dialog_error_no_route_message);
        builder.setPositiveButton(R.string.navigation_dialog_error_no_route_ok, new DialogInterface.OnClickListener() {
            @Override
            public void onClick(DialogInterface dialogInterface, int i) {
                dialogInterface.dismiss();
            }
        });
        builder.show();
    }

    private void setCameraTilt() {
        CameraPosition cameraPosition =
                new CameraPosition.Builder(mapView.getCameraController().getCameraPosition())
                        .setTilt(TILT_NAVIGATION)
                        .setZoom(ZOOM_LEVEL)
                        .build();
        mapView.getCameraController().animateCamera(cameraPosition, ANIMATION_DURATION, new MapView.CameraAnimationListener() {
            @Override
            public void onAnimationEnded(boolean b) {
                mapView.setFollowGpsPosition(true);
            }
        });
    }

    private void applyAllNavigationPreferences() {
        applyGuidanceOutputPreference();
        applyLanguagePreference();
        applyAutozoomPreference();
        applyAutozoomModePreference();
        updateRoutesOnMap();
        applyRouteGeofencePreference();
        applyRouteGeofenceAccuracyPreference();
    }

    private void applyGuidanceOutputPreference() {
        String guidanceOutputString = PreferenceUtils.getString(
                this,
                R.string.key_navigation_guidanceOutput,
                R.string.defaultValue_navigation_guidanceOutput
        );


        GuidanceOutput guidanceOutput = getGuidanceOutputFromPersistedString(
                this, guidanceOutputString
        );

        sdk.getRouteNavigator().getGuidanceSettings().setGuidanceOutput(guidanceOutput);
    }

    private void applyLanguagePreference() {
        String value = PreferenceUtils.getString(
                this,
                R.string.key_navigation_language,
                R.string.defaultValue_navigation_language
        );

        sdk.getRouteNavigator()
                .getGuidanceSettings()
                .setLocale(value);
    }

    private void applyAutozoomPreference() {
        boolean value = PreferenceUtils.getBoolean(
                this,
                R.string.key_navigation_autozoomEnabled,
                R.bool.defaultValue_navigation_autozoomEnabled
        );

        mapView.setAutoZoomEnabled(value);
    }

    private void applyAutozoomModePreference() {
        String autozoomModeString = PreferenceUtils.getString(
                this,
                R.string.key_navigation_autozoomMode,
                R.string.defaultValue_navigation_autozoomMode
        );


        AutoZoomMode autozoomMode = getAutozoomModeFromPersistedString(
                this, autozoomModeString
        );

        mapView.setAutoZoomMode(autozoomMode);
    }

    private void applyRouteGeofencePreference() {
        boolean routeGeofenceEnabled = PreferenceUtils.getBoolean(
                this,
                R.string.key_navigation_routeGeofenceEnabled,
                R.bool.defaultValue_navigation_routeGeofenceEnabled
        );
        sdk.getRouteGeofenceManager().setEnabled(routeGeofenceEnabled);
    }

    private void applyRouteGeofenceAccuracyPreference() {
        String routeGeofenceAccuracy = PreferenceUtils.getString(
                this,
                R.string.key_navigation_routeGeofenceAccuracy,
                R.string.defaultValue_navigation_routeGeofenceAccuracy
        );
        double accuracy = Double.parseDouble(routeGeofenceAccuracy);
        sdk.getRouteGeofenceManager().setAccuracy(accuracy);
    }

    // endregion

    private RouteCalculationOptions buildRouteCalculationOptions() {
        final RouteCalculationOptions options = RouteCalculationOptions.create();

        final boolean considerTraffic = PreferenceUtils.getBoolean(
                this,
                R.string.key_routing_considerTraffic,
                R.bool.defaultValue_routing_considerTraffic
        );
        options.setConsiderTraffic(considerTraffic);

        final boolean avoidTollRoads = PreferenceUtils.getBoolean(
                this,
                R.string.key_routing_avoidTollRoads,
                R.bool.defaultValue_routing_avoidTollRoads
        );
        options.setAvoidTollRoads(avoidTollRoads);

        final String routingOptimizationString = PreferenceUtils.getString(
                this,
                R.string.key_routing_optimization,
                R.string.defaultValue_routing_optimization
        );

        final RoutingOptimization routingOptimization =
                getRouteOptimizationByPreferenceString(this, routingOptimizationString);

        if (routingOptimization != null) {
            options.setOptimization(routingOptimization);
        }

        final String vehicleTypeString = PreferenceUtils.getString(
                this,
                R.string.key_routing_vehicleType,
                R.string.defaultValue_routing_vehicleType
        );

        final VehicleType vehicleType =
                getVehicleTypeByPreferenceString(this, vehicleTypeString);

        if (vehicleType != null) {
            options.setVehicleType(vehicleType);
        }

        final boolean walkingMode = PreferenceUtils.getBoolean(
                this,
                R.string.key_routing_walkingMode,
                R.bool.defaultValue_routing_walkingMode);

        options.setIsWalkingMode(walkingMode);

        return options;
    }

    private void addSdkListeners() {
        sdk.getNavigationManager().addNavigationListener(navigationListener);
        sdk.getPositionManager().addPositionListener(positionListener);
        sdk.getRouteNavigator().addRouteVisualizerRequestListener(routeVisualizerRequestListener);
        sdk.getRouteGeofenceManager().addRouteGeofenceListener(routeGeofenceListener);
    }

    private void removeSdkListeners() {
        sdk.getNavigationManager().removeNavigationListener(navigationListener);
        sdk.getPositionManager().removePositionListener(positionListener);
        sdk.getRouteNavigator().removeRouteVisualizerRequestListener(routeVisualizerRequestListener);
        sdk.getRouteGeofenceManager().removeRouteGeofenceListener(routeGeofenceListener);
    }

    private void onRouteCalculationStarted() {
        clearRoutes();

        routeCalculationDialog = new ProgressDialog(NavigationActivity.this);

        routeCalculationDialog.setTitle(R.string.dialog_wait);
        routeCalculationDialog.setMessage(getString(R.string.dialog_calculating_route));
        routeCalculationDialog.setIndeterminate(true);
        routeCalculationDialog.setCancelable(true);
        routeCalculationDialog.setCanceledOnTouchOutside(false);

        routeCalculationDialog.setButton(
                DialogInterface.BUTTON_NEGATIVE,
                getString(android.R.string.cancel),
                new DialogInterface.OnClickListener() {
                    @Override
                    public void onClick(DialogInterface dialog, int which) {
                        dialog.cancel();
                    }
                });

        routeCalculationDialog.setOnCancelListener(
                new DialogInterface.OnCancelListener() {
                    @Override
                    public void onCancel(DialogInterface dialog) {
                        if (routeCalculationSession != null) {
                            routeCalculationSession.cancel();
                        }
                    }
                });

        routeCalculationDialog.show();

        clearRoutesInfo();
        updateRouteButtons();
    }

    private void onRouteCleared() {
        updateRoutingInfoView();
        updateNavigationInfoView();
        updateRouteButtons();
        updateRouteGeofence();
    }

    @Override
    public void onCameraMoved(CameraMovementReason cameraMovementReason) {
        if (cameraMovementReason == CameraMovementReason.GESTURE) {
            mapView.setFollowGpsPosition(false);
            mapView.setFollowGpsDirection(false);
        }
    }

    private class SdkStateListener implements Sdk.StateChangeListener {
        @Override
        public void onStateChanged(Sdk.State state) {
            if (state == Sdk.State.INITIALIZED) {
                sdk.setActiveMapView(mapView);

                addSdkListeners();

                applyAllNavigationPreferences();

                speedLimitController.setNavigationManager(sdk.getNavigationManager());

                updateGpsStatusView();
            }

            updateControlsVisibility();
        }
    }

    private class PreferenceChangeCallback implements PreferenceChangeListener.Callback {
        @Override
        public void onPreferenceChanged(@StringRes int keyId) {
            switch (keyId) {
                case R.string.key_navigation_guidanceOutput:
                    applyGuidanceOutputPreference();
                    break;
                case R.string.key_navigation_language:
                    applyLanguagePreference();
                    break;
                case R.string.key_navigation_autozoomEnabled:
                    applyAutozoomPreference();
                    break;
                case R.string.key_navigation_autozoomMode:
                    applyAutozoomModePreference();
                    break;
                case R.string.key_navigation_routeColor:
                    updateRoutesOnMap();
                    break;
                case R.string.key_navigation_routeGeofenceEnabled:
                    applyRouteGeofencePreference();
                    break;
                case R.string.key_navigation_routeGeofenceAccuracy:
                    applyRouteGeofenceAccuracyPreference();
                    break;
            }
        }
    }

    private class NavigationListener implements NavigationManager.NavigationListener {
        @Override
        public void onNavigationInfoUpdated() {
            updateNavigationInfoView();
        }

        @Override
        public void onCountryChanged() {
        }

        @Override
        public void onDemoRouteStarted() {
            updateRouteButtons();
        }

        @Override
        public void onDemoRouteFinished() {
            updateRouteButtons();
        }
    }

    private class PositionListener implements PositionManager.PositionListener {
        @Override
        public void onGpsStatusChanged(GpsStatus gpsStatus) {
            Log.i("PositionListener", "onGpsStatusChanged: " + gpsStatus);

            updateGpsStatusView();
        }

        @Override
        public void onGpsPositionChanged(Location location, Position position) {
            Log.i("PositionListener", String.format(
                    "onGpsPositionChanged, raw: %s, snapped: %s", location, position.coordinates()));
        }

        @Override
        public void onExtrapolatedPositionChanged(MapCoordinates mapCoordinates) {
        }
    }

    private class RouteCalculationListener extends com.navmii.sdk.routecalculation.RouteCalculationListener {
        @Override
        public void onRouteCalculationSucceeded(RouteCalculationResult routeCalculationResult) {

            NavigationActivity.this.routeCalculationResult = routeCalculationResult;

            routes = routeCalculationResult.getSortedRoutes();
            selectedRouteIndex = !routes.isEmpty() ? 0 : -1;

            closeRouteCalculationDialog();
            updateRoutingInfoView();
            updateNavigationInfoView();
            updateRouteButtons();
            updateRoutesOnMap();
            updateRouteGeofence();
        }

        @Override
        public void onRouteCalculationFailed(
                RouteCalculationFailureInfo routeCalculationFailureInfo) {

            closeRouteCalculationDialog();

            Toast.makeText(
                    NavigationActivity.this,
                    getString(R.string.routing_error, routeCalculationFailureInfo.getErrorCode()),
                    Toast.LENGTH_SHORT).show();
        }

        @Override
        public void onNonFatalRouteCalculationErrorOccurred(
                RouteCalculationError routeCalculationError) {

            Toast.makeText(
                    NavigationActivity.this,
                    getString(R.string.routing_error, routeCalculationError),
                    Toast.LENGTH_SHORT).show();
        }
    }

    private class RouteVisualizerRequestListener
            extends com.navmii.sdk.routevisualization.RouteVisualizerRequestListener {

        @Override
        public void onRouteVisualizerRequestAvailable(
                RouteVisualizerRequest routeVisualizerRequest) {

            NavigationActivity.this.routeVisualizerRequest = routeVisualizerRequest;
            sdk.getRouteVisualizer().performRequest(routeVisualizerRequest);
        }
    }

    private class RouteGeofenceListener implements RouteGeofenceManager.RouteGeofenceListener {

        @Override
        public void onRouteGeofenceEnter(String routeName) {
            Snackbar.make(mapView,
                    "Enter to route geofence " + routeName,
                    3000
            ).show();
        }

        @Override
        public void onRouteGeofenceExit(String routeName) {
            Snackbar.make(mapView,
                    "Exit route geofence " + routeName,
                    3000
            ).show();
        }
    }
}
