package com.navmii.sdksampleapp.places;

import android.content.Intent;
import android.content.SharedPreferences;
import android.os.Bundle;
import android.preference.PreferenceManager;
import android.text.TextUtils;
import android.view.MenuItem;
import android.view.View;
import android.view.WindowManager;

import androidx.annotation.Nullable;
import androidx.appcompat.app.AppCompatActivity;

import com.google.android.material.bottomsheet.BottomSheetBehavior;
import com.navmii.sdk.Sdk;
import com.navmii.sdk.common.MapCoordinates;
import com.navmii.sdk.common.MapRectangle;
import com.navmii.sdk.common.Place;
import com.navmii.sdk.common.PoiCategory;
import com.navmii.sdk.map.CameraPosition;
import com.navmii.sdk.map.MapView;
import com.navmii.sdk.util.Util;
import com.navmii.sdksampleapp.R;
import com.navmii.sdksampleapp.sdk.SdkFragment;
import com.navmii.sdksampleapp.views.PlacesSheetView;

import java.util.ArrayList;
import java.util.List;


public class PlacesActivity extends AppCompatActivity
        implements View.OnClickListener, MapView.MapTapListener, Sdk.PoiCategoriesListener, MapView.PlaceListener,
        PlacesSheetView.Listener {

    private final Sdk sdk = Sdk.getInstance();
    private final Sdk.StateChangeListener stateChangeListener = new SdkStateListener();

    private MapView mapView;
    private View initializationProgressBar;
    private View settingsButton;
    private PlacesSheetView placesSheetView;
    private BottomSheetBehavior<View> placesSheetBehavior;
    private float cameraZoomForSelectedPlaces = 14;

    private final SharedPreferences.OnSharedPreferenceChangeListener preferenceChangeListener =
            new SharedPreferencesChangeListener();

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.activity_places);
        getWindow().addFlags(WindowManager.LayoutParams.FLAG_KEEP_SCREEN_ON);

        if (getSupportActionBar() != null) {
            getSupportActionBar().setDisplayHomeAsUpEnabled(true);
            getSupportActionBar().setDisplayShowHomeEnabled(true);
        }

        if (getFragmentManager().findFragmentByTag(SdkFragment.SDK_FRAGMENT_TAG) == null) {
            getFragmentManager()
                    .beginTransaction()
                    .add(new SdkFragment(), SdkFragment.SDK_FRAGMENT_TAG)
                    .commit();
        }

        mapView = findViewById(R.id.map_view);
        mapView.addMapTapListener(this);
        mapView.addPoiListener(this);

        initializationProgressBar = findViewById(R.id.initialization_progressbar);
        settingsButton = findViewById(R.id.settings_button);
        settingsButton.setOnClickListener(this);

        placesSheetView = findViewById(R.id.bottom_sheet);
        placesSheetView.setListener(this);
        placesSheetBehavior = BottomSheetBehavior.from(placesSheetView);
        placesSheetBehavior.setState(BottomSheetBehavior.STATE_HIDDEN);

        sdk.addPoiCategoriesListener(this);
    }

    @Override
    protected void onDestroy() {
        super.onDestroy();

        mapView.removeMapTapListener(this);
        mapView.removeMapTapListener(this);
        sdk.removePoiCategoriesListener(this);
    }

    @Override
    protected void onStart() {
        super.onStart();

        if (sdk.getState() == Sdk.State.INITIALIZED) {
            sdk.setActiveMapView(mapView);
            applyPoiCategoryVisibilityPreference();
        }

        sdk.addStateChangeListener(stateChangeListener);

        updateControlsVisibility();

        PreferenceManager.getDefaultSharedPreferences(this)
                .registerOnSharedPreferenceChangeListener(preferenceChangeListener);
    }

    @Override
    protected void onStop() {
        super.onStop();

        if (sdk.getState() == Sdk.State.INITIALIZED) {
            sdk.setActiveMapView(null);
        }

        sdk.removeStateChangeListener(stateChangeListener);

        PreferenceManager.getDefaultSharedPreferences(this)
                .unregisterOnSharedPreferenceChangeListener(preferenceChangeListener);
    }

    @Override
    public boolean onOptionsItemSelected(MenuItem item) {
        if (item.getItemId() == android.R.id.home) {
            finish();
        }

        return super.onOptionsItemSelected(item);
    }

    @Override
    public void onSingleTap(List<Place> places) {
        List<MapCoordinates> coordinates = new ArrayList<>();
        for (Place place : places) {
            coordinates.add(place.getCoordinates());
        }
        MapRectangle boundingRectangle = Util.getBoundingRectangle(
                coordinates.toArray(new MapCoordinates[0]));

        cameraZoomForSelectedPlaces = mapView.getCameraController()
                .getCameraPositionForMapRectangle(boundingRectangle, 0, 0)
                .getZoom();

        placesSheetView.setPlaces(places);
        placesSheetBehavior.setState(BottomSheetBehavior.STATE_COLLAPSED);
    }

    @Override
    public void onDoubleTap(List<Place> place) {
    }

    @Override
    public void onLongTap(List<Place> place) {
    }

    // region PlacesSheetView.Listener
    @Override
    public void loadPlaceDetails(Place place, MapView.PlaceDetailsCallback callback) {
        mapView.loadPlaceDetails(place, callback);
    }

    @Override
    public void onPlaceSelected(Place place) {
        MapView.CameraController cameraController = mapView.getCameraController();
        CameraPosition cameraPosition = new CameraPosition.Builder(cameraController.getCameraPosition())
                .setTargetLocation(place.getCoordinates())
                .setZoom(cameraZoomForSelectedPlaces)
                .build();
        cameraController.animateCamera(cameraPosition);
    }
    // endregion

    private class SdkStateListener implements Sdk.StateChangeListener {
        @Override
        public void onStateChanged(Sdk.State state) {
            if (state == Sdk.State.INITIALIZED) {
                sdk.setActiveMapView(mapView);
                applyPoiCategoryVisibilityPreference();
            }

            updateControlsVisibility();
        }
    }

    private void updateControlsVisibility() {
        initializationProgressBar.setVisibility(
                sdk.getState() == Sdk.State.INITIALIZING ? View.VISIBLE : View.GONE);

        settingsButton.setVisibility(
                sdk.getState() == Sdk.State.INITIALIZED ? View.VISIBLE : View.GONE);
    }

    @Override
    public void onClick(View view) {
        switch (view.getId()) {
            case R.id.settings_button:
                startPlacesPreferenceActivity();
                break;
        }
    }

    private class SharedPreferencesChangeListener implements
            SharedPreferences.OnSharedPreferenceChangeListener {

        @Override
        public void onSharedPreferenceChanged(SharedPreferences sharedPreferences, String key) {
            String poiCategoryName = getPoiCategoryNameByPreferenceKey(key);
            PoiCategory category = findPoiCategoryByName(poiCategoryName);

            if (category != null) {
                applyPoiCategoryVisibilityPreference(category);
            }
        }
    }

    @Nullable
    private PoiCategory findPoiCategoryByName(String poiCategoryName) {
        for (PoiCategory category : sdk.getPoiCategories()) {
            if (TextUtils.equals(poiCategoryName, category.getName())) {
                return category;
            }
        }

        return null;
    }

    private void startPlacesPreferenceActivity() {
        final Intent startIntent = PlacesPreferenceActivity
                .createStartIntent(this, getPoiCategoryNames());

        startActivity(startIntent);
    }

    private ArrayList<String> getPoiCategoryNames() {
        PoiCategory[] poiCategories = sdk.getPoiCategories();

        ArrayList<String> poiCategoryNames = new ArrayList<>(poiCategories.length);

        for (PoiCategory category : poiCategories) {
            poiCategoryNames.add(category.getName());
        }

        return poiCategoryNames;
    }

    private static final String POI_CATEGORY_PREFERENCE_KEY_PREFIX = "key_poiCategory.";

    @Nullable
    public static String getPoiCategoryNameByPreferenceKey(String key) {
        if (key.startsWith(POI_CATEGORY_PREFERENCE_KEY_PREFIX) &&
                key.length() > POI_CATEGORY_PREFERENCE_KEY_PREFIX.length()) {

            return key.substring(POI_CATEGORY_PREFERENCE_KEY_PREFIX.length());
        }

        return null;
    }

    public static String getPoiCategoryPreferenceKey(String poiCategoryName) {
        return POI_CATEGORY_PREFERENCE_KEY_PREFIX + poiCategoryName;
    }

    public static boolean getPoiCategoryPreferenceDefaultValue(String poiCategoryName) {
        // maybe something changes later depending on category
        return true;
    }

    private void applyPoiCategoryVisibilityPreference(PoiCategory poiCategory) {
        final String key = getPoiCategoryPreferenceKey(poiCategory.getName());
        final boolean defaultValue = getPoiCategoryPreferenceDefaultValue(poiCategory.getName());

        //noinspection ConstantConditions
        boolean shouldShowOnMap = PreferenceManager
                .getDefaultSharedPreferences(this)
                .getBoolean(key, defaultValue);

        poiCategory.setShouldDisplayOnMap(shouldShowOnMap);
    }

    private void applyPoiCategoryVisibilityPreference() {
        for (PoiCategory poiCategory : sdk.getPoiCategories()) {
            applyPoiCategoryVisibilityPreference(poiCategory);
        }
    }

    @Override
    public void onSingleTap(MapCoordinates mapCoordinates) {
        placesSheetBehavior.setState(BottomSheetBehavior.STATE_HIDDEN);
    }

    @Override
    public void onDoubleTap(MapCoordinates mapCoordinates) {

    }

    @Override
    public void onLongTap(MapCoordinates mapCoordinates) {

    }

    @Override
    public void onPoiCategoriesUpdated() {
        applyPoiCategoryVisibilityPreference();
    }
}