package com.navmii.sdksampleapp.poisearch;

import android.app.ProgressDialog;
import android.content.Intent;
import android.graphics.PointF;
import android.os.Bundle;
import android.os.Handler;
import android.util.Log;
import android.view.MenuItem;
import android.view.View;
import android.view.WindowManager;
import android.widget.SearchView;
import android.widget.Toast;

import androidx.annotation.Nullable;
import androidx.appcompat.app.AppCompatActivity;
import androidx.appcompat.widget.Toolbar;

import com.navmii.sdk.Sdk;
import com.navmii.sdk.common.MapCoordinates;
import com.navmii.sdk.common.Place;
import com.navmii.sdk.common.PoiCategory;
import com.navmii.sdk.map.CameraPosition;
import com.navmii.sdk.map.GeoMarker;
import com.navmii.sdk.map.MapView;
import com.navmii.sdk.positioning.Position;
import com.navmii.sdk.search.PoiRequest;
import com.navmii.sdk.search.Request;
import com.navmii.sdk.search.RequestError;
import com.navmii.sdksampleapp.CommonPreferenceActivity;
import com.navmii.sdksampleapp.PreferenceUtils;
import com.navmii.sdksampleapp.R;
import com.navmii.sdksampleapp.SearchResultsActivity;
import com.navmii.sdksampleapp.sdk.AddressInfo;
import com.navmii.sdksampleapp.sdk.PlaceInfo;
import com.navmii.sdksampleapp.sdk.SdkFragment;

import java.io.File;
import java.util.ArrayList;
import java.util.List;

public class PoiSearchActivity extends AppCompatActivity
        implements View.OnClickListener, SearchView.OnQueryTextListener {

    private static final int SEARCH_RESULTS_REQUEST_CODE = 1;

    private final Sdk sdk = Sdk.getInstance();
    private final Sdk.StateChangeListener stateChangeListener = new SdkStateListener();

    private MapView mapView;
    private View initializationProgressBar;
    private File flag;
    private SearchView searchView;

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.activity_poi_search);
        getWindow().addFlags(WindowManager.LayoutParams.FLAG_KEEP_SCREEN_ON);

        final Toolbar toolbar = findViewById(R.id.toolbar);
        searchView = toolbar.findViewById(R.id.search_bar);
        setSupportActionBar(toolbar);

        if (getSupportActionBar() != null) {
            getSupportActionBar().setDisplayHomeAsUpEnabled(true);
            getSupportActionBar().setDisplayShowHomeEnabled(true);
        }

        if (getFragmentManager().findFragmentByTag(SdkFragment.SDK_FRAGMENT_TAG) == null) {
            getFragmentManager()
                    .beginTransaction()
                    .add(new SdkFragment(), SdkFragment.SDK_FRAGMENT_TAG)
                    .commit();
        }

        mapView = findViewById(R.id.map_view);
        searchView.setOnQueryTextListener(this);
        initializationProgressBar = findViewById(R.id.initialization_progressbar);
        findViewById(R.id.settings_button).setOnClickListener(this);
    }

    @Override
    protected void onStart() {
        super.onStart();

        if (sdk.getState() == Sdk.State.INITIALIZED) {
            sdk.setActiveMapView(mapView);
        }

        sdk.addStateChangeListener(stateChangeListener);

        updateControlsVisibility();
    }

    @Override
    protected void onStop() {
        super.onStop();

        if (sdk.getState() == Sdk.State.INITIALIZED) {
            sdk.setActiveMapView(null);
        }

        sdk.removeStateChangeListener(stateChangeListener);
    }

    @Override
    public boolean onOptionsItemSelected(MenuItem item) {
        switch (item.getItemId()) {
            case android.R.id.home:
                finish();
                return true;
        }

        return super.onOptionsItemSelected(item);
    }

    private class SdkStateListener implements Sdk.StateChangeListener {
        @Override
        public void onStateChanged(Sdk.State state) {
            if (state == Sdk.State.INITIALIZED) {
                sdk.setActiveMapView(mapView);
            }

            updateControlsVisibility();
        }
    }

    private void updateControlsVisibility() {
        initializationProgressBar.setVisibility(
                sdk.getState() == Sdk.State.INITIALIZING ? View.VISIBLE : View.GONE);

        searchView.setVisibility(
                sdk.getState() == Sdk.State.INITIALIZED ? View.VISIBLE : View.GONE);

        findViewById(R.id.settings_button).setVisibility(
                sdk.getState() == Sdk.State.INITIALIZED ? View.VISIBLE : View.GONE);
    }

    @Override
    public void onClick(View view) {
        switch (view.getId()) {
            case R.id.settings_button:
                startPlacesSearchPreferenceActivity();
                break;
        }
    }

    private void startPlacesSearch(String query) {
        if (sdk.getState() != Sdk.State.INITIALIZED) {
            return;
        }

        final String maxResultCountString = PreferenceUtils.getString(
                this,
                R.string.key_poiSearch_maxResultCount,
                R.string.defaultValue_poiSearch_maxResultCount
        );

        final int maxResultCount = Integer.parseInt(maxResultCountString);

        final Position currentPosition = sdk.getPositionManager().getCurrentPosition();
        if (currentPosition == null) {
            Toast.makeText(this, R.string.no_current_position, Toast.LENGTH_SHORT).show();
            return;
        }

        final PoiRequest.Builder builder = sdk.getSearchManager()
                .createPoiRequest(query, currentPosition.coordinates())
                .setMaxResultCount(maxResultCount);

        final boolean searchWithinVisibleArea = PreferenceUtils.getBoolean(
                this,
                R.string.key_poiSearch_searchWithinVisibleArea,
                R.bool.defaultValue_poiSearch_searchWithinVisibleArea);

        if (searchWithinVisibleArea) {
            MapView.CameraController cameraController = mapView.getCameraController();
            builder.setViewBox(cameraController.getMapRectangleForCameraPosition(
                    cameraController.getCameraPosition()));
        } else {
            final int searchRadius = PreferenceUtils.getInteger(
                    this,
                    R.string.key_poiSearch_searchRadius,
                    R.integer.defaultValue_poiSearch_searchRadius);
            builder.setSearchRadius(searchRadius);
        }

        final Request poiRequest = builder.build();

        final ProgressDialog progressDialog = ProgressDialog.show(
                this,
                getString(R.string.dialog_loading),
                getString(R.string.dialog_wait),
                true
        );

        poiRequest.start(new Request.CompletionListener() {
            @Override
            public void onRequestCompleted(Request request, List<Place> results) {
                progressDialog.dismiss();
                startSearchResultsActivity(currentPosition.coordinates(), results);
            }

            @Override
            public void onRequestFailed(Request request, RequestError error) {
                Log.i("PoiRequest", "Request failed: " + error);
            }
        });
    }

    @Override
    public boolean onQueryTextSubmit(String query) {
        startPlacesSearch(query);
        return false;
    }

    @Override
    public boolean onQueryTextChange(String newText) {
        return false;
    }

    private PlaceInfo convertPlaceToPlaceInfo(Place place) {
        AddressInfo address = new AddressInfo(place.getAddress(), place.getCoordinates());
        int primaryCategory = place.getPoiData().getPrimaryCategory();

        String category = "";
        PoiCategory poiCategory = sdk.getPoiCategory(primaryCategory);

        if (poiCategory != null) {
            category = poiCategory.getName();
        }

        return new PlaceInfo(place.getName(), category, address);
    }

    private void startSearchResultsActivity(MapCoordinates searchCoordinates, List<Place> results) {
        final ArrayList<PlaceInfo> places = new ArrayList<>();
        for (Place result : results) {
            places.add(convertPlaceToPlaceInfo(result));
        }

        final Intent startIntent = SearchResultsActivity.createStartIntent(
                this, searchCoordinates, places);
        startActivityForResult(startIntent, SEARCH_RESULTS_REQUEST_CODE);
    }

    @Override
    protected void onActivityResult(int requestCode, int resultCode, final Intent data) {
        super.onActivityResult(requestCode, resultCode, data);

        if (requestCode == SEARCH_RESULTS_REQUEST_CODE && resultCode == RESULT_OK) {

            // onActivityResult may be called before onStart,
            // when it's prohibited to used some of MapView functionality.
            new Handler().post(new Runnable() {
                @Override
                public void run() {
                    MapCoordinates mapCoordinates = data.getParcelableExtra(
                            SearchResultsActivity.EXTRA_RESULT_MAP_COORDINATES);

                    moveCamera(mapCoordinates);
                    mapView.removeAllGeoObjects();
                    mapView.setFollowGpsPosition(false);
                    addMarker(mapCoordinates);
                }
            });
        }
    }

    private void moveCamera(MapCoordinates mapCoordinates) {
        CameraPosition cameraPosition =
                new CameraPosition.Builder(mapView.getCameraController().getCameraPosition())
                        .setTargetLocation(mapCoordinates)
                        .build();

        mapView.getCameraController().animateCamera(cameraPosition);
    }

    private void addMarker(MapCoordinates mapCoordinates) {
        GeoMarker marker = new GeoMarker(mapCoordinates, getFlagFile(), getFlagAnchorPoint());
        marker.setClickable(false);
        mapView.addGeoObject(marker);
    }

    @Nullable
    private File getFlagFile() {
        if (flag == null) {
            flag = new File(generateFlagPath());
        }

        return flag;
    }

    private String generateFlagPath() {
        String fileDirs = this.getFilesDir().getAbsolutePath();
        return fileDirs + File.separator + "assets" + File.separator + "flag_green.png";
    }

    private PointF getFlagAnchorPoint() {
        return new PointF(0.3f, 0.9f);
    }

    private void startPlacesSearchPreferenceActivity() {
        Intent startIntent = CommonPreferenceActivity.createStartIntent(
                this,
                R.xml.preferences_poi_search,
                R.string.title_poiSearchPreferences
        );

        startActivity(startIntent);
    }
}
