package com.navmii.sdksampleapp.reversegeocoding;

import android.app.ProgressDialog;
import android.content.DialogInterface;
import android.content.Intent;
import android.graphics.PointF;
import android.os.Bundle;
import android.os.Handler;
import android.view.MenuItem;
import android.view.View;
import android.view.WindowManager;
import android.widget.SearchView;

import androidx.annotation.Nullable;
import androidx.appcompat.app.AlertDialog;
import androidx.appcompat.app.AppCompatActivity;
import androidx.appcompat.widget.Toolbar;

import com.navmii.sdk.Sdk;
import com.navmii.sdk.common.MapCoordinates;
import com.navmii.sdk.common.Place;
import com.navmii.sdk.common.PoiCategory;
import com.navmii.sdk.map.CameraPosition;
import com.navmii.sdk.map.GeoMarker;
import com.navmii.sdk.map.MapView;
import com.navmii.sdk.positioning.Position;
import com.navmii.sdk.search.Request;
import com.navmii.sdk.search.RequestError;
import com.navmii.sdksampleapp.R;
import com.navmii.sdksampleapp.SearchResultsActivity;
import com.navmii.sdksampleapp.sdk.AddressInfo;
import com.navmii.sdksampleapp.sdk.PlaceInfo;
import com.navmii.sdksampleapp.sdk.SdkFragment;

import java.io.File;
import java.util.ArrayList;
import java.util.List;

public class ReverseGeocodingActivity extends AppCompatActivity
        implements SearchView.OnQueryTextListener {

    private static final int SEARCH_RESULTS_REQUEST_CODE = 1;

    private final Sdk sdk = Sdk.getInstance();
    private final Sdk.StateChangeListener stateChangeListener = new SdkStateListener();
    private final MapView.MapTapListener mapTapListener = new MapTapListener();

    private MapView mapView;
    private View initializationProgressBar;
    private File flag;
    private SearchView searchViewLat;
    private SearchView searchViewLon;

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.activity_reverse_geocoding);
        getWindow().addFlags(WindowManager.LayoutParams.FLAG_KEEP_SCREEN_ON);

        final Toolbar toolbar = findViewById(R.id.toolbar);
        searchViewLat = toolbar.findViewById(R.id.search_bar_lat);
        searchViewLon = toolbar.findViewById(R.id.search_bar_lon);
        setSupportActionBar(toolbar);

        if (getSupportActionBar() != null) {
            getSupportActionBar().setDisplayHomeAsUpEnabled(true);
            getSupportActionBar().setDisplayShowHomeEnabled(true);
        }

        if (getFragmentManager().findFragmentByTag(SdkFragment.SDK_FRAGMENT_TAG) == null) {
            getFragmentManager()
                    .beginTransaction()
                    .add(new SdkFragment(), SdkFragment.SDK_FRAGMENT_TAG)
                    .commit();
        }

        mapView = findViewById(R.id.map_view);
        searchViewLat.setOnQueryTextListener(this);
        searchViewLon.setOnQueryTextListener(this);
        initializationProgressBar = findViewById(R.id.initialization_progressbar);
    }

    @Override
    protected void onStart() {
        super.onStart();

        if (sdk.getState() == Sdk.State.INITIALIZED) {
            sdk.setActiveMapView(mapView);
        }

        sdk.addStateChangeListener(stateChangeListener);

        mapView.addMapTapListener(mapTapListener);

        updateControlsVisibility();
    }

    @Override
    protected void onStop() {
        super.onStop();

        if (sdk.getState() == Sdk.State.INITIALIZED) {
            sdk.setActiveMapView(null);
        }

        mapView.removeMapTapListener(mapTapListener);

        sdk.removeStateChangeListener(stateChangeListener);
    }

    @Override
    public boolean onOptionsItemSelected(MenuItem item) {
        switch (item.getItemId()) {
            case android.R.id.home:
                finish();
                return true;
        }

        return super.onOptionsItemSelected(item);
    }

    private class SdkStateListener implements Sdk.StateChangeListener {
        @Override
        public void onStateChanged(Sdk.State state) {
            if (state == Sdk.State.INITIALIZED) {
                sdk.setActiveMapView(mapView);
            }

            updateControlsVisibility();
        }
    }

    private class MapTapListener implements MapView.MapTapListener {
        @Override
        public void onSingleTap(MapCoordinates mapCoordinates) {
            mapView.removeAllGeoObjects();
            addMarker(mapCoordinates);
            startReverseGeoCodingSearch(mapCoordinates);
        }

        @Override
        public void onDoubleTap(MapCoordinates mapCoordinates) {
        }

        @Override
        public void onLongTap(MapCoordinates mapCoordinates) {
        }
    }

    private void updateControlsVisibility() {
        initializationProgressBar.setVisibility(
                sdk.getState() == Sdk.State.INITIALIZING ? View.VISIBLE : View.GONE);

        searchViewLat.setVisibility(
                sdk.getState() == Sdk.State.INITIALIZED ? View.VISIBLE : View.GONE);
        searchViewLon.setVisibility(
                sdk.getState() == Sdk.State.INITIALIZED ? View.VISIBLE : View.GONE);
    }

    private void startReverseGeoCodingSearch(MapCoordinates searchCoordinates) {
        if (sdk.getState() != Sdk.State.INITIALIZED) {
            return;
        }

        Request placesSearch = sdk
                .getSearchManager()
                .createReverseGeocodingRequest(searchCoordinates)
                .build();

        final ProgressDialog progressDialog = ProgressDialog.show(
                this,
                getString(R.string.dialog_loading),
                getString(R.string.dialog_wait),
                true
        );

        placesSearch.start(new Request.CompletionListener() {
            @Override
            public void onRequestCompleted(Request request, List<Place> results) {
                progressDialog.dismiss();
                startSearchResultsActivity(searchCoordinates, results);
            }

            @Override
            public void onRequestFailed(Request request, RequestError error) {
                progressDialog.dismiss();
                showErrorDialog();
            }
        });
    }

    private PlaceInfo convertPlaceToPlaceInfo(Place place) {
        AddressInfo address = new AddressInfo(place.getAddress(), place.getCoordinates());
        int primaryCategory = place.getPoiData().getPrimaryCategory();

        String category = "";
        PoiCategory poiCategory = sdk.getPoiCategory(primaryCategory);

        if (poiCategory != null) {
            category = poiCategory.getName();
        }

        return new PlaceInfo(place.getName(), category, address);
    }

    private void startSearchResultsActivity(MapCoordinates searchCoordinates, List<Place> results) {
        final ArrayList<PlaceInfo> places = new ArrayList<>();
        for (Place result : results) {
            places.add(convertPlaceToPlaceInfo(result));
        }

        final Intent startIntent = SearchResultsActivity.createStartIntent(
                this, searchCoordinates, places);
        startActivityForResult(startIntent, SEARCH_RESULTS_REQUEST_CODE);
    }

    @Override
    protected void onActivityResult(int requestCode, int resultCode, final Intent data) {
        super.onActivityResult(requestCode, resultCode, data);

        if (requestCode == SEARCH_RESULTS_REQUEST_CODE && resultCode == RESULT_OK) {

            // onActivityResult may be called before onStart,
            // when it's prohibited to used some of MapView functionality.
            new Handler().post(new Runnable() {
                @Override
                public void run() {
                    MapCoordinates mapCoordinates = data.getParcelableExtra(
                            SearchResultsActivity.EXTRA_RESULT_MAP_COORDINATES);

                    mapView.setFollowGpsPosition(false);
                    moveCamera(mapCoordinates);
                }
            });
        }
    }

    @Override
    public boolean onQueryTextSubmit(String query) {
        double lat;
        double lon;
        try {
            lat = Double.valueOf(searchViewLat.getQuery().toString());
            lon = Double.valueOf(searchViewLon.getQuery().toString());
        } catch (NumberFormatException e) {
            showErrorDialog();
            return false;
        }
        MapCoordinates mapCoordinates = new MapCoordinates(lat, lon);
        mapView.removeAllGeoObjects();
        addMarker(mapCoordinates);
        startReverseGeoCodingSearch(mapCoordinates );
        return false;
    }

    @Override
    public boolean onQueryTextChange(String newText) {
        return false;
    }

    private void addMarker(MapCoordinates mapCoordinates) {
        GeoMarker marker = new GeoMarker(mapCoordinates, getFlagFile(), getFlagAnchorPoint());
        marker.setClickable(false);
        mapView.addGeoObject(marker);
    }

    @Nullable
    private File getFlagFile() {
        if (flag == null) {
            flag = new File(generateFlagPath());
        }

        return flag;
    }

    private String generateFlagPath() {
        String fileDirs = this.getFilesDir().getAbsolutePath();
        return fileDirs + File.separator + "assets" + File.separator + "flag_green.png";
    }

    private PointF getFlagAnchorPoint() {
        return new PointF(0.3f, 0.9f);
    }

    private void showErrorDialog() {
        AlertDialog.Builder builder = new AlertDialog.Builder(this);
        builder.setTitle(R.string.dialog_error_coordinates_title);
        builder.setMessage(R.string.dialog_error_coordinates_message);
        builder.setPositiveButton(R.string.ok, new DialogInterface.OnClickListener() {
            @Override
            public void onClick(DialogInterface dialogInterface, int i) {
                dialogInterface.dismiss();
            }
        });
        builder.show();
    }

    private void moveCamera(MapCoordinates mapCoordinates) {
        CameraPosition cameraPosition =
                new CameraPosition.Builder(mapView.getCameraController().getCameraPosition())
                        .setTargetLocation(mapCoordinates)
                        .build();
        mapView.getCameraController().animateCamera(cameraPosition);
    }
}
