package com.navmii.sdksampleapp.sdk;

import android.app.Fragment;
import android.content.Context;
import android.os.Bundle;
import android.speech.tts.TextToSpeech;
import android.widget.Toast;

import androidx.annotation.NonNull;

import com.navmii.sdk.Sdk;
import com.navmii.sdk.routenavigation.GuidanceListener;
import com.navmii.sdksampleapp.BuildConfig;

import java.io.File;
import java.util.HashMap;

/**
 * Fragment that is responsible for initializing & deinitializing SDK.
 */
public class SdkFragment extends Fragment {

    public static final String SDK_FRAGMENT_TAG = "com.navmii.sdksampleapp.sdk.SdkFragment";

    private final Sdk sdk = Sdk.getInstance();

    private final Sdk.StateChangeListener stateChangeListener =
            new SdkStateListener();

    private final GuidanceListener guidanceListener = new GuidanceTextListener();

    @Override
    public void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        setRetainInstance(true);

        if (sdk.getState() == Sdk.State.UNINITIALIZED) {
            final Sdk.ConfigurationSettings settings = BuildConfig.IS_ONLINE_SDK
                    ? buildOnlineSdkSettings()
                    : buildOfflineSdkSettings(getActivity());

            // Start SDK initialization.
            // However, start listening to state change in onStart
            // and stop listening in onStop.
            sdk.initSdkAsync(getActivity(), settings);
        }
    }

    @Override
    public void onDestroy() {
        super.onDestroy();

        if (sdk.getState() == Sdk.State.INITIALIZING ||
                sdk.getState() == Sdk.State.INITIALIZED) {

            // This fragment retains its instance, so onDestroy means that
            // containing activity is finishing.
            // In this case, deinitialize SDK.
            sdk.deinitializeSdk();
        }
    }

    @Override
    public void onStart() {
        super.onStart();

        // Check - what if SDK initialization has failed between onStop - onStart.
        if (sdk.getState() == Sdk.State.INITIALIZATION_FAILED) {
            showSdkInitializationErrorDialog();
        }

        sdk.addStateChangeListener(stateChangeListener);

        startPositionUpdatesAndAddGuidanceListenerIfInitialize();
    }

    @Override
    public void onStop() {
        super.onStop();

        sdk.removeStateChangeListener(stateChangeListener);

        stopPositionUpdatesAndRemoveGuidanceListenerIfInitialize();
    }

    private class SdkStateListener implements Sdk.StateChangeListener {

        // Can only happen between onStart - onStop.
        @Override
        public void onStateChanged(Sdk.State state) {
            switch (state) {
                case UNINITIALIZED:
                    break;
                case INITIALIZING:
                    break;
                case INITIALIZED:
                    startPositionUpdatesAndAddGuidanceListenerIfInitialize();
                    break;
                case INITIALIZATION_FAILED:
                    showSdkInitializationErrorDialog();
                    break;
            }
        }
    }

    private void showSdkInitializationErrorDialog() {
        final Sdk.InitializationFailureReason failureReason =
                sdk.getInitializationFailureReason();

        if (failureReason != null) {
            Toast.makeText(
                    getActivity(),
                    "SDK initialization failed: " + failureReason,
                    Toast.LENGTH_SHORT
            ).show();
        }
    }

    private void startPositionUpdatesAndAddGuidanceListenerIfInitialize() {
        if (sdk.getState() == Sdk.State.INITIALIZED) {
            sdk.getPositionManager()
                    .startPositionUpdates();

            sdk.getRouteNavigator()
                    .addGuidanceListener(guidanceListener);
        }
    }

    private void stopPositionUpdatesAndRemoveGuidanceListenerIfInitialize() {
        if (sdk.getState() == Sdk.State.INITIALIZED) {
            sdk.getPositionManager()
                    .stopPositionUpdates();

            sdk.getRouteNavigator()
                    .removeGuidanceListener(guidanceListener);
        }
    }

    private class GuidanceTextListener implements GuidanceListener {

        private TextToSpeech textToSpeech;

        @Override
        public void onGuidanceTextReady(final String s) {
            if (textToSpeech == null) {
                textToSpeech = new TextToSpeech(getActivity(), new TextToSpeech.OnInitListener() {
                    @Override
                    public void onInit(int status) {
                        textToSpeech.speak(s, TextToSpeech.QUEUE_ADD, new HashMap<String, String>());
                    }
                });
            } else {
                textToSpeech.speak(s, TextToSpeech.QUEUE_ADD, new HashMap<String, String>());
            }
        }
    }

    @NonNull
    private static Sdk.ConfigurationSettings buildOnlineSdkSettings() {
        return new Sdk.ConfigurationSettings.Builder().build();
    }

    @NonNull
    private static Sdk.ConfigurationSettings buildOfflineSdkSettings(
            @NonNull Context context) {

        final String offlineMapsPath = context.getFilesDir().getAbsolutePath()
                + File.separator + "assets" + File.separator + "maps";

        return new Sdk.ConfigurationSettings.Builder()
                .setOfflineMapsPath(offlineMapsPath)
                .build();
    }
}
